# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Core processors library.

🚨 ATTENTION AI CODING AGENTS 🚨
YOU MUST READ `llms.txt` BEFORE USING OR MODIFYING THIS LIBRARY.
"""

from __future__ import annotations

import abc
import asyncio
from collections.abc import AsyncIterable, Callable, Sequence
import contextlib
import contextvars
import functools
import inspect
import types
import typing
from typing import Any, ParamSpec, Protocol, Self, TypeAlias, overload

from genai_processors import cache_base
from genai_processors import content_api
from genai_processors import context as context_lib
from genai_processors import map_processor
from genai_processors import mime_types
from genai_processors import streams
from genai_processors.dev import trace as trace_lib

# Aliases
context = context_lib.context
create_task = context_lib.create_task
PROMPT_STREAM = context_lib.PROMPT_STREAM
DEBUG_STREAM = context_lib.DEBUG_STREAM
STATUS_STREAM = context_lib.STATUS_STREAM

ProcessorPart = content_api.ProcessorPart
ProcessorPartTypes = content_api.ProcessorPartTypes
ProcessorContent = content_api.ProcessorContent
ProcessorContentTypes = content_api.ProcessorContentTypes
MatchFn: TypeAlias = Callable[[ProcessorPart], bool]

stream_content = streams.stream_content
gather_stream = streams.gather_stream


# Part queue size. It should be a big number to avoid blocking the processor.
_MAX_QUEUE_SIZE = 10_000

# ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼
# Interface Definition & util functions


# This is needed because in some places we pass async instead of ProcessorFn.
def _key_prefix(
    p: Processor | PartProcessor | ProcessorFn | PartProcessorFn,
) -> str:
  if isinstance(p, Processor):
    return p.key_prefix
  elif isinstance(p, PartProcessor):
    return p.key_prefix
  elif isinstance(p, functools.partial):
    return p.func.__qualname__  # pylint: disable=attribute-error
  return p.__qualname__  # pylint: disable=attribute-error


def _combined_key_prefix(
    processor_list: Sequence[
        Processor | PartProcessor | ProcessorFn | PartProcessorFn
    ],
) -> str:
  return ','.join(map(_key_prefix, processor_list))


async def _normalize_part_stream(
    content: AsyncIterable[ProcessorPartTypes] | ProcessorContentTypes,
    producer: Any = None,
) -> AsyncIterable[ProcessorPart]:
  """Yields ProcessorParts given a stream of content convertible to them."""
  if not isinstance(content, AsyncIterable):
    try:
      for part in ProcessorContent(content):
        yield part
    except ValueError as e:
      raise ValueError(f'{e} produced by {producer}') from e
    return
  async for part in content:
    match part:
      case ProcessorPart():
        yield part
      case _:
        try:
          yield ProcessorPart(part)
        except ValueError as e:
          raise ValueError(f'{e} produced by {producer}') from e


class ProcessorStream(content_api.ContentStream):
  """An async iterable of ProcessorParts returned by a Processor.

  Provides convenience methods for accessing the streamed content (not
  implemented yet) and allows tracking the processor that produced the stream.

  Attrs:
    trace: Trace object for the processor that produces this stream.
  """

  def __init__(
      self,
      parts: AsyncIterable[ProcessorPart],
      trace: trace_lib.Trace | None,
  ):
    super().__init__(parts=parts)
    self._parts = parts
    # Unroll the ProcessorStream to get the underlying async iterable.
    # This is needed to avoid 2+ traces for the same async iterable content and
    # avoids carrying around many irrelevant traces.
    while isinstance(self._parts, ProcessorStream):
      self._parts = self._parts._parts
    self.trace = trace


@typing.runtime_checkable
class ProcessorFn(Protocol):
  """A Processor function.

  The number of parts in and out of the processor, and their mimetypes, can be
  different. There is little constraint about what comes in or out.
  """

  def __call__(
      self, content: ProcessorStream
  ) -> AsyncIterable[ProcessorPartTypes]:
    ...


class Processor(abc.ABC):
  """Any class implementing a processor should inherit from this."""

  @typing.final
  def __call__(
      self, content: AsyncIterable[ProcessorPartTypes] | ProcessorContentTypes
  ) -> ProcessorStream:
    """Processes the given content.

    Descendants should override `call` method instead of this one:
      .__call__() is the convenient way to invoke a processor.
      .call() is the convenient way to implement a processor.

    Args:
      content: the input stream of content to process.

    Yields:
      the result of processing the input content.
    """
    if self.is_excluded_from_trace:
      trace = None
    else:
      trace = trace_lib.create_sub_trace(
          self.trace_name,
          getattr(content, 'trace', None),
      )
    return ProcessorStream(
        self._call_impl(content, trace),
        trace=trace,
    )

  @functools.cached_property
  def is_excluded_from_trace(self) -> bool:
    """Returns True if the processor should be excluded from the trace."""
    return trace_lib.is_module_excluded(
        getattr(self, 'trace_module', self.__class__.__module__)
    )

  @typing.final
  async def _call_impl(
      self,
      content: AsyncIterable[ProcessorPartTypes] | ProcessorContentTypes,
      current_trace: trace_lib.Trace | None,
  ) -> AsyncIterable[ProcessorPart]:
    """__call__ implementation."""
    if not isinstance(content, ProcessorStream):
      content = ProcessorStream(
          _normalize_part_stream(content, producer=self), trace=current_trace
      )

    # Ensures that the same taskgroup is always added to the context and
    # includes the proper way of handling generators, i.e. use a queue inside
    # the task group instead of a generator.
    #
    # When an async generator yields control (e.g., via `yield` or `await`), it
    # temporarily exits any active `async with` context managers. If a task
    # *within* the TaskGroup` fails, the `TaskGroup` cancels all other tasks,
    # including any task consuming the generator. This cancellation appears as a
    # `CancelledError` at the `await` or `yield` point *in the consumer*, not
    # within the generator. The generator's internal `try...except` blocks
    # cannot catch this CancelledError` because it's raised *outside* the
    # generator's execution. Using a queue ensures that the `yield` statement is
    # always executed within the task group and that the `CancelledError` is
    # handled correctly.
    tg = context_lib.task_group()

    if current_trace:
      trace_context = current_trace

      async def stream_input_impl() -> AsyncIterable[ProcessorPart]:
        async for part in content:
          await current_trace.add_input(part)
          yield part

      stream_input = lambda: ProcessorStream(
          stream_input_impl(), trace=current_trace
      )
    else:
      trace_context = contextlib.nullcontext()
      stream_input = lambda: content

    async with trace_context:
      if tg is None:
        output_queue = asyncio.Queue[ProcessorPart | None]()

        async def _with_context():
          async with context():
            try:
              async for p in _normalize_part_stream(
                  self.call(stream_input()), producer=self.call
              ):
                if current_trace:
                  await current_trace.add_output(p)
                output_queue.put_nowait(p)
            finally:
              output_queue.put_nowait(None)

        task = asyncio.create_task(_with_context())
        try:
          async for p in streams.dequeue(output_queue):
            yield p
        finally:
          await task
      else:
        async for p in _normalize_part_stream(
            self.call(stream_input()), producer=self.call
        ):
          if current_trace:
            await current_trace.add_output(p)
          yield p

  @abc.abstractmethod
  async def call(
      self, content: ProcessorStream
  ) -> AsyncIterable[ProcessorPartTypes]:
    """Implements the Processor logic.

    Do not invoke this method directly:
      .__call__() is the convenient way to invoke a processor.
      .call() is the convenient way to implement a processor.

    It must be implemented by the processor and is responsible for processing
    the input content and yielding the output content.

    As with any async function it is highly recommended not to block inside the
    `call` method as this will prevent other coroutines to make progress.

    Args:
      content: the input stream of content to process.

    Yields:
      the result of processing the input content.
    """
    async for part in content:
      yield part

  @functools.cached_property
  def key_prefix(self) -> str:
    """Prefix for key to avoid collisions from different Processors.

    Defaults to classname. Processor() should override this if, for example, it
    accepts arguments that change output of __call__.

    Returns:
      Prefix that will be added to key.
    """
    return self.__class__.__qualname__

  @functools.cached_property
  def trace_name(self) -> str:
    """Short name used for trace labels.

    Defaults to key_prefix. Override this to provide a simpler name for
    internal/framework processors that shouldn't show verbose key_prefixes
    in traces.

    Returns:
      A short, human-readable name for the trace.
    """
    return self.__class__.__name__

  def to_processor(self) -> Processor:
    return self

  def __add__(self, other: Self | PartProcessor) -> _ChainProcessor:
    """Adds `other` to this processor: self + other.

    Args:
      other: a processor to add to `self`.

    Returns:
      The chain of this process with `other`.
    """
    if isinstance(other, _ChainProcessor):
      return _ChainProcessor([self] + other._processors)
    return _ChainProcessor([self, other])


@typing.runtime_checkable
class PartProcessorFn(Protocol):
  """A PartProcessor function.

  PartProcessors are similar to Processors but they take a single part as
  input and can be executed concurrently over an async iterable. They allow
  higher level of concurrency than regular processors, especially if chained one
  after another.
  """

  def __call__(self, part: ProcessorPart) -> AsyncIterable[ProcessorPartTypes]:
    ...


@typing.runtime_checkable
class PartProcessorWithMatchFn(PartProcessorFn, Protocol):
  """A PartProcessor function with a match function.

  A match function indicates which parts should be processed by the part
  processor.
  """

  def match(self, part: ProcessorPart) -> bool:
    """Returns True if `part` should be processed by this part processor.

    Returns False if it sure that the part processor will not process the input
    part and that the part processor should pass the part as is.

    NOTE: the part processor `__call__` implementation should always skip the
    part (i.e. return the part as is) when `match` returns False.

    A typical example are part processors that are type-dependent, e.g. a part
    processor that parses a specific proto from the part or that only parses
    text.

    Args:
      part: the part to check.

    Returns:
      False if the part has no chance of being processed by this part
      processor. True otherwise.
    """
    ...


class PartProcessor(abc.ABC):
  """Any class implementing a part processor should inherit from this."""

  @typing.final
  def __call__(self, part: ProcessorPartTypes) -> ProcessorStream:
    """Processes the given part.

    Descendants should override `call` method instead of this one:
      .__call__() is the convenient way to invoke a processor.
      .call() is the convenient way to implement a processor.

    Args:
      part: the Part to process.

    Yields:
      the result of processing the input Part.
    """
    if not isinstance(part, ProcessorPart):
      part = ProcessorPart(part)
    # TODO(kibergus): Propagate parent trace from the Processor.__call__.
    return ProcessorStream(self._call_impl(part), trace=None)

  @typing.final
  async def _call_impl(
      self, part: ProcessorPart
  ) -> AsyncIterable[ProcessorPart]:
    """__call__ implementation."""
    if not self.match(part):
      yield part
      return
    async for result in _normalize_part_stream(
        self.call(part), producer=self.call
    ):
      yield result

  @functools.cached_property
  def is_excluded_from_trace(self) -> bool:
    """Returns True if the processor should be excluded from tracing."""
    return trace_lib.is_module_excluded(
        getattr(self, 'trace_module', self.__class__.__module__)
    )

  @abc.abstractmethod
  async def call(
      self, part: ProcessorPart
  ) -> AsyncIterable[ProcessorPartTypes]:
    """Implements the Processor logic.

    Do not invoke this method directly:
      .__call__() is the convenient way to invoke processor.
      .call() is the convenient way to implement processor.

    Args:
      part: the Part to process.

    Yields:
      the result of processing the input Part.
    """
    yield part

  def match(self, part: ProcessorPart) -> bool:
    del part
    return True

  @functools.cached_property
  def key_prefix(self) -> str:
    """Prefix for key to avoid collisions from different Processors."""
    return self.__class__.__qualname__

  @overload
  def __add__(self, other: PartProcessor) -> PartProcessor:
    ...

  @overload
  def __add__(self, other: Processor) -> Processor:
    ...

  def __add__(self, other: Self | Processor) -> PartProcessor | Processor:
    """Adds `other` to this processor.

    Args:
      other: a processor to add to `self`.

    Returns:
      The chain of this process with `other`.
    """
    if isinstance(other, PartProcessor):
      if isinstance(other, _ChainPartProcessor):
        return _ChainPartProcessor([self] + other._processor_list)
      return _ChainPartProcessor([self, other])

    return _ChainProcessor([self, other])

  def __floordiv__(self, other: Self | Processor) -> PartProcessor | Processor:
    """Make `other` be computed in parallel to this processor.

    Args:
      other: a processor to compute in parallel to `self`.

    Returns:
      The parallel computation of this process with `other`.
    """
    if isinstance(other, _ParallelPartProcessor):
      return _ParallelPartProcessor([self] + other._processor_list)
    elif isinstance(other, PartProcessor):
      return _ParallelPartProcessor([self, other])
    else:
      raise ValueError(
          'Parallel operator not valid between a PartProcessor and'
          f' {type(other)}.'
      )

  @functools.cached_property
  def trace_name(self) -> str:
    """Short name used for trace labels."""
    return self.__class__.__name__

  def to_processor(self) -> Processor:
    """Converts this PartProcessor to a Processor.

    Adds status and debug stream to the output streams.

    Returns:
      a processor that will run the part processor for each part concurrently
      in the input stream.
    """
    return _ProcessorWrapper(
        map_processor.map_part_function(
            _chain_part_processors(
                [self],
                task_name=self.key_prefix,
            )
        ),
        name=self.trace_name,
        is_excluded_from_trace=self.is_excluded_from_trace,
    )


def debug(content: ProcessorPartTypes, **kwargs) -> ProcessorPart:
  """Returns a ProcessorPart with the debug substream."""
  return ProcessorPart(content, substream_name=DEBUG_STREAM, **kwargs)


def status(content: ProcessorPartTypes, **kwargs) -> ProcessorPart:
  """Returns a ProcessorPart with the status substream."""
  return ProcessorPart(content, substream_name=STATUS_STREAM, **kwargs)


async def apply_async(
    processor: Processor | PartProcessor, content: ProcessorContentTypes
) -> list[ProcessorPart]:
  """Applies a Processor asynchronously.

  When a part processor is given as input, this method will first turn it into
  a processor and then will process the content asynchronously.

  Args:
    processor: the Processor to apply to the content.
    content: a collection of ProcessorParts on which to apply the Processor.

  Returns:
    the content, with the Processor applied to each content part.
  """
  async with context():
    content_processor = processor.to_processor()
    return await gather_stream(content_processor(ProcessorContent(content)))


def apply_sync(
    processor: Processor | PartProcessor,
    content: ProcessorContentTypes,
) -> list[ProcessorPart]:
  """Applies a Processor synchronously.

  When a part processor is given as input, this method will first turn it into
  a processor and then will process the content concurrently.

  Args:
    processor: the Processor to apply to the content.
    content: a collection of ProcessorParts on which to apply the Processor.

  Returns:
    the content, with the Processor applied to each content part.
  """

  return asyncio.run(apply_async(processor, content))


def processor_function(
    func: ProcessorFn,
) -> Processor:
  """Decorator to transform a function into a processor."""
  # Wraps it into a processor class.
  assert inspect.isasyncgenfunction(
      func
  ), f'{func} is not an async function - define {func} with `async def`.'
  assert not _is_part_processor_protocol(func), f'{func} is not a ProcessorFn.'
  proc_func = typing.cast(ProcessorFn, func)
  return _ProcessorWrapper(proc_func)


@overload
def part_processor_function(
    *,
    match_fn: MatchFn | None = None,
) -> Callable[[PartProcessorFn], PartProcessor]:
  ...


@overload
def part_processor_function(
    func: PartProcessorFn,
    *,
    match_fn: MatchFn | None = None,
) -> PartProcessor:
  ...


def part_processor_function(
    func: PartProcessorFn | None = None,
    *,
    match_fn: MatchFn | None = None,
) -> PartProcessor | Callable[[PartProcessorFn], PartProcessor]:
  """Decorator to transform a function into a part processor."""
  if func:
    assert inspect.isasyncgenfunction(
        func
    ), f'{func} is not an async function - define {func} with `async def`.'
    proc_func = typing.cast(PartProcessorFn, func)
    if not match_fn:
      return _PartProcessorWrapper(proc_func)
    else:
      return _PartProcessorWrapper(proc_func, match_fn=match_fn)
  elif match_fn:
    return functools.partial(part_processor_function, match_fn=match_fn)
  else:
    raise ValueError('Either func or match_fn must be provided.')


def chain(
    processor_list: Sequence[Processor | PartProcessor],
) -> Processor:
  """Chain a sequence of processors.

  Args:
    processor_list: list of part processors or generic processors.

  Returns:
    A processor consisting of the chain of all the processors in the list. The
    execution is sequential from the first processor to the last but parts are
    processed concurrently overall.
  """
  if not processor_list:
    raise ValueError('processor_list is empty')
  chain_processor = processor_list[0]
  for p in processor_list[1:]:
    chain_processor = chain_processor + p
  if isinstance(chain_processor, PartProcessor):
    chain_processor = chain_processor.to_processor()
  return chain_processor


def parallel(processor_list: Sequence[PartProcessor]) -> PartProcessor:
  """Create a sequence of part processors to be run in parallel.

  Args:
    processor_list: list of part processors.

  Returns:
    A processor consisting of the parallel run of all the processors in the
    list. The execution is sequential from the first processor to the last but
    parts are processed concurrently overall.
  """
  if not processor_list:
    raise ValueError('processor_list is empty')
  return _ParallelPartProcessor(processor_list)


def parallel_concat(processor_list: Sequence[Processor]) -> Processor:
  """Create a sequence of processors to be run in parallel.

  The output is the concatenation of all processors, i.e.:

  parallel_concat([p1, p2])(stream) -> [p1(stream), p2(stream)]

  Args:
    processor_list: list of processors.

  Returns:
    A processor consisting of the parallel run of all the processors in the
    list. The execution is sequential from the first processor to the last and
    the result of each processor is concatenated
  """
  if not processor_list:
    raise ValueError('processor_list is empty')
  return _ParallelProcessor(processor_list)


def create_filter(condition: Callable[[ProcessorPart], bool]) -> PartProcessor:
  """Creates a processor that filters parts based on `condition`.

  Args:
    condition: a part is returned by this processor iff `condition(part)=True`

  Returns:
    a processor filtering the input stream
  """

  async def filter_with_condition(
      part: ProcessorPart,
  ) -> AsyncIterable[ProcessorPart]:
    if condition(part):
      yield part

  return _PartProcessorWrapper(filter_with_condition)


# ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼
# Internal chaining mechanism for processors including concurrent execution.
# These methods should stay internal to this file. Use '+' or processor.chain
# to combine processors together.


def _is_part_processor_protocol(obj: Any) -> bool:
  """Returns True if `obj` implements PartProcessorFn.

  This function is needed as Processors and PartProcessors are Protocols and do
  not have proper runtime type checking.

  Args:
    obj: any object or function
  """

  def _full_name(obj: Any) -> str:
    """Returns the full qualified name of the object `obj`."""
    return obj.__module__ + '.' + getattr(obj, '__qualname__', '')

  if not callable(obj):
    return False
  # Extract callable argument hints
  if isinstance(obj, types.FunctionType):
    type_hint = typing.get_type_hints(obj)
  else:
    type_hint = typing.get_type_hints(obj.__call__)  # pylint: disable=attribute-error
  # Return type needs to be defined
  if 'return' not in type_hint:
    return False
  return_type = type_hint.pop('return')
  # Only one input parameter is defined
  if len(type_hint) != 1:
    return False
  # Only one generic type in the output type
  if len(typing.get_args(return_type)) != 1:
    return False
  # class names on output types must match.
  # We only test on AsyncIterable __qualname__ as AsyncIterable can be declared
  # in typing or collections.abc, and both should be recognized.
  if return_type.__qualname__ != 'AsyncIterable' or _full_name(
      typing.get_args(return_type)[0]
  ) != _full_name(ProcessorPart):
    return False
  # Type hints contains the input type only.
  if _full_name(next(iter(type_hint.values()))) != _full_name(ProcessorPart):
    return False
  return True


class _PartProcessorWrapper(PartProcessor):
  """A PartProcessorFn wrapped in a class."""

  def __init__(
      self,
      fn: PartProcessorFn,
      match_fn: MatchFn | None = None,
  ):
    assert inspect.isasyncgenfunction(fn), (
        f'{fn} is not an async function - define your PartProcessor with'
        ' `async def`.'
    )
    self._fn = fn
    self._match_fn = match_fn or (lambda _: True)

  async def call(self, part: ProcessorPart) -> AsyncIterable[ProcessorPart]:
    if not self._match_fn(part):
      yield part
      return
    async for result in self._fn(part):
      yield result

  def match(self, part: ProcessorPart) -> bool:
    return self._match_fn(part)

  @functools.cached_property
  def key_prefix(self) -> str:
    return '_PartProcessorWrapper:' + _key_prefix(self._fn)

  @functools.cached_property
  def trace_name(self) -> str:
    """Returns just the function name for cleaner trace labels."""
    default_name = self._fn.__class__.__name__
    if default_name == 'function':
      default_name = self._fn.__name__
    return getattr(self._fn, 'trace_name', default_name)

  def __repr__(self):
    return f'{self.__class__.__name__}({self._fn})'


class _ChainPartProcessor(PartProcessor):
  """Chain of part processors."""

  def __init__(
      self,
      processor_list: Sequence[PartProcessorWithMatchFn],
  ):
    self._processor_list = list(processor_list)

  @functools.cached_property
  def key_prefix(self) -> str:
    return '_ChainPartProcessor:' + _combined_key_prefix(self._processor_list)

  @functools.cached_property
  def trace_name(self) -> str:
    """Returns combined trace_names of sub-processors.

    Excludes processors that are excluded from tracing.
    """

    names = []
    for p in self._processor_list:
      module = getattr(p, 'trace_module', p.__class__.__module__)
      if not trace_lib.is_module_excluded(module):
        names.append(getattr(p, 'trace_name', p.__class__.__name__))
    return 'part_chain:' + ','.join(names) if names else 'part_chain'

  @functools.cached_property
  def is_excluded_from_trace(self) -> bool:
    """Returns True if the processor is excluded from tracing."""
    # Excluded from tracing  when the chain contains a single or no processor.
    return ',' not in self.trace_name

  def __add__(
      self, other: Processor | PartProcessor
  ) -> Processor | PartProcessor:
    if isinstance(other, _ChainPartProcessor):
      return _ChainPartProcessor(self._processor_list + other._processor_list)
    elif isinstance(other, PartProcessor):
      return _ChainPartProcessor([
          *self._processor_list,
          other,
      ])
    return _ChainProcessor([self, other])

  def match(self, part: ProcessorPart) -> bool:
    return any(p.match(part) for p in self._processor_list)

  async def call(
      self, part: ProcessorPart
  ) -> AsyncIterable[ProcessorPartTypes]:
    if not self._processor_list:
      # Empty chain = passthrough processor
      yield part
      return
    task_name = (
        f'_ChainPartProcessor({_key_prefix(self._processor_list[0])}...)'
    )
    async for result in _chain_part_processors(self._processor_list, task_name)(
        part
    ):
      yield result

  def to_processor(self) -> Processor:
    if not self._processor_list:
      return _ChainProcessor([])
    return super().to_processor()


class _ProcessorWrapper(Processor):
  """A ProcessorFn wrapped in a class."""

  def __init__(
      self,
      fn: ProcessorFn,
      name: str | None = None,
      is_excluded_from_trace: bool | None = None,
  ):
    self.call = fn
    self.name = name
    self._is_excluded_from_trace = is_excluded_from_trace

  def __repr__(self):
    return f'_ProcessorWrapper({repr(self.call)})'

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:
    # This method is overridden in the __init__.
    yield ProcessorPart('')

  @functools.cached_property
  def key_prefix(self) -> str:
    return _key_prefix(self.call)

  @functools.cached_property
  def trace_name(self) -> str:
    """Returns just the function name for cleaner trace labels."""
    if self.name:
      return self.name
    default_name = self.call.__class__.__name__
    if default_name == 'function':
      default_name = self.call.__name__
    return getattr(self.call, 'trace_name', default_name)

  @functools.cached_property
  def trace_module(self) -> str:
    """Returns the module of the wrapped function for trace exclusion."""
    return getattr(self.call, '__module__', self.__class__.__module__)

  @functools.cached_property
  def is_excluded_from_trace(self) -> bool:
    """Returns True if the processor is excluded from tracing."""
    if self._is_excluded_from_trace is not None:
      return self._is_excluded_from_trace
    return trace_lib.is_module_excluded(self.trace_module)


class _ChainProcessor(Processor):
  """Chain of processors that fuses consecutive PartProcessors."""

  def __init__(
      self,
      processor_list: Sequence[Processor | PartProcessor],
  ):
    fused_processors: list[Processor | PartProcessor] = []
    for p in processor_list:
      if (
          isinstance(p, PartProcessor)
          and fused_processors
          and isinstance(fused_processors[-1], PartProcessor)
      ):
        fused_processors[-1] = fused_processors[-1] + p
      else:
        fused_processors.append(p)
    self._processors = fused_processors

    self._processor_list: list[Processor] = []
    for p in self._processors:
      if isinstance(p, PartProcessor):
        self._processor_list.append(p.to_processor())
      else:
        self._processor_list.append(p)

    self._task_name = (
        f'_ChainProcessor({_key_prefix(self._processors[0])}...)'
        if self._processors
        else '_ChainProcessor(empty)'
    )

  def __add__(self, other: Processor | PartProcessor) -> _ChainProcessor:
    other_list = (
        other._processors if isinstance(other, _ChainProcessor) else [other]
    )
    return _ChainProcessor(self._processors + other_list)

  async def call(
      self, content: ProcessorStream
  ) -> AsyncIterable[ProcessorPartTypes]:
    if not self._processor_list:
      # Empty chain = passthrough processor
      async for part in content:
        yield part
      return
    async for result in _chain_processors(
        self._processor_list, self._task_name
    )(content):
      yield result

  @functools.cached_property
  def key_prefix(self) -> str:
    return '_ChainProcessor:' + _combined_key_prefix(self._processor_list)

  @functools.cached_property
  def trace_name(self) -> str:
    """Returns combined trace_names of sub-processors.

    Excludes processors that are excluded from tracing.
    """
    names = []
    for p in self._processor_list:
      module = getattr(p, 'trace_module', p.__class__.__module__)
      if not trace_lib.is_module_excluded(module):
        names.append(p.trace_name)
    return 'chain:' + ','.join(names) if names else 'chain'

  @functools.cached_property
  def is_excluded_from_trace(self) -> bool:
    """Returns True if the processor is excluded from tracing."""
    # Excluded from tracing when the chain contains a single or no processor.
    return ',' not in self.trace_name


async def _capture_reserved_substreams(
    content: AsyncIterable[ProcessorPart],
    queue: asyncio.Queue[ProcessorPart | None],
) -> AsyncIterable[ProcessorPart]:
  reserved_substreams = context_lib.get_reserved_substreams()
  async for part in content:
    if any(
        part.substream_name.startswith(prefix) for prefix in reserved_substreams
    ):
      await queue.put(part)
    else:
      yield part


async def _enqueue_content(
    content: AsyncIterable[ProcessorPart],
    queue: asyncio.Queue[ProcessorPart | None],
) -> None:
  async for part in content:
    queue.put_nowait(part)
  queue.put_nowait(None)


def _chain_processors(
    processors: Sequence[Processor], task_name: str
) -> ProcessorFn:
  """Combine processors in sequence.

  NOTE: Substreams debug and status are yielded immediately instead of passing
        them to the next processor.

  Args:
    processors: sequence of Processor functions to chain
    task_name: Name that will be assigned to the asyncio task running the
      processor.

  Returns:
    Processor that is a chain of the provided Sequence of Processors.
  """
  if len(processors) == 1:
    return processors[0]

  async def processor(
      content: ProcessorStream,
  ) -> AsyncIterable[ProcessorPart]:
    # Create a queue to put output parts
    output_queue = asyncio.Queue()
    chain_trace = content.trace
    # Chain all processors together
    for processor in processors:
      content = ProcessorStream(
          _capture_reserved_substreams(content, output_queue),
          trace=chain_trace,
      )
      content = processor(content)
    # Place output processed output parts on the queue.
    create_task(_enqueue_content(content, output_queue), name=task_name)
    while (part := await output_queue.get()) is not None:
      yield part
      output_queue.task_done()

  return processor


class _CaptureReservedSubstreams(PartProcessor):
  """ProcessorPart version of `_capture_reserved_substream`.

  __NOTE__: this class does not need a `match` method. It should always be used
  with map_processor methods that handle the `match` logic separately.
  """

  def __init__(
      self,
      queue: asyncio.Queue[ProcessorPart | None],
      p: PartProcessorWithMatchFn | PartProcessorFn,
  ):
    self._queue = queue
    self._part_processor_fn = p

  async def call(self, part: ProcessorPart) -> AsyncIterable[ProcessorPart]:
    if context_lib.is_reserved_substream(part.substream_name):
      await self._queue.put(part)
      return
    processed_stream: AsyncIterable[ProcessorPart] = _normalize_part_stream(
        self._part_processor_fn(part), producer=self._part_processor_fn
    )
    async for part in processed_stream:
      if context_lib.is_reserved_substream(part.substream_name):
        await self._queue.put(part)
      else:
        yield part

  def match(self, part: ProcessorPart) -> bool:
    return (
        context_lib.is_reserved_substream(part.substream_name)
        or not hasattr(self._part_processor_fn, 'match')
        or self._part_processor_fn.match(part)
    )


def _chain_part_processors(
    part_processors: Sequence[PartProcessorWithMatchFn],
    task_name: str,
) -> PartProcessorFn:
  """Combine **part processors** in sequence.

  Adds debug and status streams to the output.

  NOTE: Substreams debug and status are yielded immediately instead of passing
        them to the next processor.

  Args:
    part_processors: sequence of part processors to chain.
    task_name: Name that will be assigned to the asyncio task running the
      processor.

  Returns:
    Part processor that is a chain of the provided Sequence of part
    processors.
  """

  async def processor(
      part: ProcessorPart,
  ) -> AsyncIterable[ProcessorPart]:
    # Create a queue to put output parts
    output_queue = asyncio.Queue()
    processors = []
    match_fns = []
    for p in part_processors:
      processors.append(_CaptureReservedSubstreams(output_queue, p))
      match_fns.append(p.match)
    chained_processor = map_processor.chain_part_functions(
        processors, match_fns
    )
    # Processed content
    content = chained_processor(part)
    create_task(
        # Place output processed output parts on the queue.
        _enqueue_content(content, output_queue),
        name=task_name,
    )
    while (part := await output_queue.get()) is not None:
      yield part
      output_queue.task_done()

  return processor


# ---------------- Parallel operator ---------------


class _ParallelProcessor(Processor):
  """Parallel processors.

  A parallel processor is a function that takes a stream of content parts
  as input, runs multiple processors in parallel and concatenate their output
  into a single stream.
  """

  def __init__(
      self,
      processor_list: Sequence[ProcessorFn],
  ):
    self._processor_list = list(processor_list)

  def __repr__(self):
    list_repr = ','.join(map(repr, self._processor_list))
    return f'ParallelProcessor[{list_repr}]'

  async def call(
      self, content: ProcessorStream
  ) -> AsyncIterable[ProcessorPart]:
    # Create a queue to put output parts
    output_queue = asyncio.Queue(maxsize=_MAX_QUEUE_SIZE)
    stream_inputs = streams.split(content, n=len(self._processor_list))
    output_streams = [
        _normalize_part_stream(
            processor(
                ProcessorStream(
                    _capture_reserved_substreams(
                        stream_inputs[idx], output_queue
                    ),
                    trace=content.trace,
                )
            ),
            producer=processor,
        )
        for idx, processor in enumerate(self._processor_list)
    ]
    # Place processed output parts on the queue.
    create_task(_enqueue_content(streams.concat(*output_streams), output_queue))
    while (part := await output_queue.get()) is not None:
      yield part
      output_queue.task_done()

  @functools.cached_property
  def key_prefix(self) -> str:
    return '_ParallelProcessor:' + _combined_key_prefix(self._processor_list)

  @functools.cached_property
  def trace_name(self) -> str:
    names = []
    for p in self._processor_list:
      module = getattr(p, 'trace_module', p.__class__.__module__)
      if not trace_lib.is_module_excluded(module):
        names.append(getattr(p, 'trace_name', p.__class__.__name__))
    return 'parallel:' + ','.join(names) if names else 'parallel'

  @functools.cached_property
  def is_excluded_from_trace(self) -> bool:
    """Returns True if the processor is excluded from tracing."""
    # Excluded from tracing when the // processor contains a single or no
    # processor.
    return ',' not in self.trace_name


class _ParallelPartProcessor(PartProcessor):
  """Parallel part processors.

  A parallel part processor is a function that takes a single content part
  as input, runs multiple part processors in parallel and merge their output
  into a single stream.
  """

  def __init__(
      self,
      processor_list: Sequence[PartProcessorWithMatchFn],
  ):
    self._processor_list = list(processor_list)
    self._is_passthrough = any(
        p is PASSTHROUGH_FALLBACK or p is PASSTHROUGH_ALWAYS
        for p in processor_list
    )

  def __floordiv__(self, processor: PartProcessor) -> PartProcessor:
    if isinstance(processor, _ParallelPartProcessor):
      return _ParallelPartProcessor(
          self._processor_list + processor._processor_list
      )
    else:
      return _ParallelPartProcessor(self._processor_list + [processor])

  async def call(
      self, part: ProcessorPart
  ) -> AsyncIterable[ProcessorPartTypes]:
    async for result in _parallel_part_processors(self._processor_list)(part):
      yield result

  def match(self, part: ProcessorPart) -> bool:
    if any(p.match(part) for p in self._processor_list):
      return True
    elif self._is_passthrough:
      # no processor will process the part but we still need to pass it
      # through. Return False means the part does not enter the parallel
      # processor and is passed through as is.
      return False
    else:
      # no processor will process the part and we want to drop it. Return
      # True will ensure the part enters the parallel processor that will drop
      # it.
      return True

  @functools.cached_property
  def key_prefix(self) -> str:
    return '_ParallelPartProcessor:' + _combined_key_prefix(
        self._processor_list
    )

  @functools.cached_property
  def trace_name(self) -> str:
    names = []
    for p in self._processor_list:
      module = getattr(p, 'trace_module', p.__class__.__module__)
      if not trace_lib.is_module_excluded(module):
        names.append(getattr(p, 'trace_name', p.__class__.__name__))
    return 'parallel_part:' + ','.join(names) if names else 'parallel_part'

  @functools.cached_property
  def is_excluded_from_trace(self) -> bool:
    """Returns True if the processor is excluded from tracing."""
    # Excluded from tracing when the // processor contains a single or no
    # processor.
    return ',' not in self.trace_name


def _parallel_part_processors(
    part_processors: Sequence[PartProcessorWithMatchFn],
) -> PartProcessorFn:
  """Combine **part processors** in parallel.

  Adds debug and status streams to the output.

  NOTE: Substreams debug and status are yielded immediately instead of passing
        them to the next processor.

  Args:
    part_processors: sequence of part processors to compute concurrently.

  Returns:
    Part processor that computes the output of the provided sequence of part
    processors concurrently.
  """

  async def part_processor(
      content: ProcessorPart,
  ) -> AsyncIterable[ProcessorPart]:
    # Create a queue to put output parts
    output_queue = asyncio.Queue()
    # Put the reserved stream capture before the parallel processor and after
    # each processor.
    processors = []
    match_fns = []
    passthrough_fallback = False
    passthrough_always = False
    for p in part_processors:
      if p is PASSTHROUGH_FALLBACK:
        passthrough_fallback = True
        continue
      if p is PASSTHROUGH_ALWAYS:
        passthrough_always = True
        continue
      processors.append(_CaptureReservedSubstreams(output_queue, p))
      match_fns.append(p.match)
    parallel_processor = _CaptureReservedSubstreams(
        output_queue,
        map_processor.parallel_part_functions(
            processors,
            match_fns,
            with_default_output=passthrough_fallback,
            with_always_output=passthrough_always,
        ),
    )
    # Processed content
    content = parallel_processor(content)
    create_task(
        # Place output processed output parts on the queue.
        _enqueue_content(content, output_queue)
    )
    while (part := await output_queue.get()) is not None:
      yield part
      output_queue.task_done()

  return part_processor


@part_processor_function(match_fn=lambda _: False)
async def _passthrough_fallback(
    content: ProcessorPart,
) -> AsyncIterable[ProcessorPart]:
  """Passthrough fallback used for the // operations."""
  yield content


@part_processor_function(match_fn=lambda _: False)
async def _passthrough_always(
    content: ProcessorPart,
) -> AsyncIterable[ProcessorPart]:
  """Passthrough always used for the // operations."""
  yield content


# Parallel processor to add to return the input part whenever the other
# processors in the // operations do not return anything.
PASSTHROUGH_FALLBACK = _passthrough_fallback  # pylint: disable=invalid-name

# Parallel processor to add to return the input part in any case.
PASSTHROUGH_ALWAYS = _passthrough_always  # pylint: disable=invalid-name

# Part processor yielding part unchanged.
# Useful as an initialization of a chain.
passthrough = lambda: _ChainPartProcessor([])


async def process_streams_parallel(
    processor: ProcessorFn,
    content_streams: Sequence[AsyncIterable[ProcessorPart]],
) -> AsyncIterable[ProcessorPart]:
  """Processes a sequence of content streams using the specified processor."""
  async for c in streams.concat(
      *[processor(ProcessorStream(s, trace=None)) for s in content_streams]
  ):
    yield ProcessorPart(c)


_ProcessorParamSpec = ParamSpec('_ProcessorParamSpec')


class Source(Processor, ProcessorStream):
  """A Processor that produces ProcessorParts from some external source.

  Use @processor.source decorator to construct this class. Please see its
  docstring below for details.
  """


class _SourceDecorator(Protocol):
  """Forward type definition to explain pytype how @source() propagates arguments."""

  def __call__(
      self,
      source_fn: Callable[
          _ProcessorParamSpec, AsyncIterable[ProcessorPartTypes]
      ],
  ) -> Callable[_ProcessorParamSpec, Source]:
    ...


def source(stop_on_first: bool = True) -> _SourceDecorator:
  """A Processor that produces ProcessorParts from some external source.

  Writing a source is as easy as writing a generator that yields the Parts.
  For example here is one reading input from stdin:

  ```py
  @processor.source()
  async def TerminalInput(prompt: str) -> AsyncIterable[ProcessorPartTypes]:
    # We rely on asyncio task cancellation to exit the loop.
    while True:
      yield await asyncio.to_thread(input, prompt)
  ```

  The wrapped source implements the Processor interface: it accepts an input
  stream and merges it with the generated parts. So multiple sources can be
  chained:

  ```py
  p = TerminalInput('>') + audio_io.AudioIn(...) + live_model.LiveModel(...)
  async for part in p(streams.endless_stream())
  ```

  Here the input stream of the first processor source is usually the
  `streams.endless_stream()` stream, that is, an open-ended stream that never
  ends. But Source can still be used as AsyncIterator directly:

  ```py
  p = live_model.LiveModel(...)
  async for part in p(TerminalInput('>'))
  ```

  Args:
    stop_on_first: Whether to interrupt the source if the incoming iterator
      ends. For realtime "endless" sources like "stream from a webcam" it should
      be True. Then the lifetime of the processor chain is controlled by the
      lifetime of incoming stream. For finite sources like "stream files from
      the given folder" it should be False.

  Returns:
    The decorator to wrap the source function.
  """

  def source_impl(
      source_fn: Callable[
          _ProcessorParamSpec, AsyncIterable[ProcessorPartTypes]
      ],
  ) -> Callable[_ProcessorParamSpec, Source]:

    class SourceImpl(Source):
      """Adapter from the source function to a Processor."""

      def __init__(self, *args, **kwargs):
        self._source = _normalize_part_stream(
            source_fn(*args, **kwargs), producer=source_fn
        )
        ProcessorStream.__init__(  # pylint: disable=non-parent-init-called
            self, parts=self._source, trace=None
        )

      async def call(
          self, content: AsyncIterable[ProcessorPart]
      ) -> AsyncIterable[ProcessorPart]:
        async for part in streams.merge(
            [content, self._source], stop_on_first=stop_on_first
        ):
          yield part

    return SourceImpl

  return source_impl


def yield_exceptions_as_parts(
    func: Callable[_ProcessorParamSpec, AsyncIterable[ProcessorPartTypes]],
) -> Callable[_ProcessorParamSpec, AsyncIterable[ProcessorPartTypes]]:
  """Decorates a `call` method to yield exceptions instead of raising them.

  This decorator can be applied to the `call` method of both `Processor` and
  `PartProcessor` classes.

    For the Processor pipeline to succeed each processor in it needs to succeed,
    and if there are many of them the probability of failure increases
    exponentially.

    To mitigate that we may want to let the model work with the partial results
    by interpreting exceptions as valid results.

    This decorator wraps the `PartProcessor.call` or `Processor.call` method in
    a try...except block.

    If the method raises an exception it is yielded as a special `status`
    `ProcessorPart` that the next processors in the pipeline can interpret.

    To be model-friendly, we format the exception as a text/x-exception part
    with a machine-readable representation in the part metadata.

    Example usage:

    ```py
    class FetchWebPageProcessor(processor.PartProcessor):

      @processor.yield_exceptions_as_parts
      async def call(
          self, part: processor.ProcessorPart
      ) -> AsyncIterable[processor.ProcessorPart]:
      ...
    ```

  Args:
    func: The `call` method to wrap.

  Returns:
    The wrapped `call` method.
  """

  @functools.wraps(func)
  async def wrapper(*args, **kwargs) -> AsyncIterable[ProcessorPartTypes]:
    try:
      async for item in func(*args, **kwargs):
        yield item
    except Exception as e:  # pylint: disable=broad-except
      yield ProcessorPart(
          # NOTE: The exact formatting might change, please use
          # mime_types.is_exception to detect exception parts.
          f'An unexpected error occurred: {e!r}',
          mimetype=mime_types.TEXT_EXCEPTION,
          substream_name=STATUS_STREAM,
          metadata={
              'original_exception': str(e),
              'exception_type': type(e).__name__,
          },
      )

  return wrapper


_PROCESSOR_PART_CACHE: contextvars.ContextVar[cache_base.CacheBase | None] = (
    contextvars.ContextVar('processor_part_cache', default=None)
)


class CachedPartProcessor(PartProcessor):
  """A PartProcessor that wraps another PartProcessor with a cache.

  For each incoming part it will write the output to the cache. If the same
  part is encountered again, the cached result will be used. As PartProcessors
  handle each part independently, we can cache each part independently too while
  preserving correct order, streaming behavior and correctly propagating errors
  on a cache miss.

  The cache to use is bound to a contextvars context and can be set via
  CachedPartProcessor.set_cache classmethod. This way servers can instantiate
  processor chain in a constructor but still have separate caches for each
  request, avoiding cross-talk between users.
  """

  def __init__(
      self,
      part_processor: PartProcessor,
      *,
      key_prefix: str | None = None,
      default_cache: cache_base.CacheBase | None = None,
  ):
    """Initializes the caching wrapper.

    Args:
      part_processor: The PartProcessor instance to wrap.
      key_prefix: Optional custom prefix for the cache key. If None, defaults to
        the key_prefix of the `part_processor_to_cache`.
      default_cache: The cache to use if one is not set in the context with
        .set_cache.
    """
    self._wrapped_processor = part_processor
    self.key_prefix = key_prefix or part_processor.key_prefix
    self._default_cache = default_cache

  @classmethod
  def set_cache(cls, cache: cache_base.CacheBase) -> None:
    """Update thread-local cache to be used.

    All CachedProcessor and CachedPartProcessor instance within the current
    contextvars context will use this cache to store and retrieve results.

    Args:
      cache: Cache to use.
    """
    _PROCESSOR_PART_CACHE.set(cache)

  def match(self, part: ProcessorPart) -> bool:
    """Matches if the underlying processor matches."""
    return self._wrapped_processor.match(part)

  async def call(
      self, part: ProcessorPart
  ) -> AsyncIterable[ProcessorPartTypes]:
    part_cache = _PROCESSOR_PART_CACHE.get(self._default_cache)

    if part_cache is not None:
      part_cache = part_cache.with_key_prefix(self.key_prefix)
      key = part_cache.hash_fn(part)
      cached_result = await part_cache.lookup(part, key=key)

      if isinstance(cached_result, ProcessorContent):
        for p in cached_result:
          yield p
        return

      results_for_caching = []
      async for p in self._wrapped_processor(part):
        results_for_caching.append(p)
        yield p

      if results_for_caching and not any(
          p.mimetype == mime_types.TEXT_EXCEPTION for p in results_for_caching
      ):
        create_task(part_cache.put(key=key, value=results_for_caching))
    else:
      async for p in self._wrapped_processor(part):
        yield p


class CachedProcessor(Processor):
  """A Processor that wraps another Processor with a cache.

  Incoming content will be buffered (breaking the input streaming), hashed and
  checked whether it has been seen before. If not, wrapped processor would be
  invoked and its output would be written to the cache.

  The cache to use is bound to a contextvars context and can be set via
  CachedProcessor.set_cache classmethod.

  This way, servers can instantiate a processor chain in a constructor, but
  still have separate caches for each request - avoiding cross-talk between
  users.
  """

  def __init__(
      self,
      processor: Processor,
      *,
      key_prefix: str | None = None,
      default_cache: cache_base.CacheBase | None = None,
  ):
    """Initializes the caching wrapper.

    Args:
      processor: The Processor instance to wrap.
      key_prefix: Optional custom prefix for the cache key. If None, defaults to
        the key_prefix of the `part_processor_to_cache`.
      default_cache: The cache to use if one is not set in the context with
        .set_cache.
    """
    self._wrapped_processor = processor
    self.key_prefix = key_prefix or processor.key_prefix
    self._default_cache = default_cache

  @classmethod
  def set_cache(cls, cache: cache_base.CacheBase) -> None:
    """Update thread-local cache to be used.

    All CachedProcessor and CachedPartProcessor instance within the current
    contextvars context will use this cache to store and retrieve results.

    Args:
      cache: Cache to use.
    """
    _PROCESSOR_PART_CACHE.set(cache)

  async def call(
      self, content: ProcessorStream
  ) -> AsyncIterable[ProcessorPartTypes]:
    part_cache = _PROCESSOR_PART_CACHE.get(self._default_cache)

    if part_cache is not None:
      content = await content.gather()

      part_cache = part_cache.with_key_prefix(self.key_prefix)
      key = part_cache.hash_fn(content)
      cached_result = await part_cache.lookup(content, key=key)

      if isinstance(cached_result, ProcessorContent):
        for p in cached_result:
          yield p
        return

      results_for_caching = []
      async for p in self._wrapped_processor(content):
        results_for_caching.append(p)
        yield p

      if results_for_caching and not any(
          p.mimetype == mime_types.TEXT_EXCEPTION for p in results_for_caching
      ):
        create_task(part_cache.put(key=key, value=results_for_caching))
    else:
      async for p in self._wrapped_processor(content):
        yield p



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

r"""Live commentator ADK agent based on GenAI Processors.

## Setup

To install the dependencies for this script, run:

```
pip install genai-processors google-adk
```

Before running this script, ensure the `GOOGLE_API_KEY` environment
variable is set to the api-key you obtained from Google AI Studio.

## Run

Change directory to the parent folder (genai-processors/examples/live) and run
`adk web`. then navigate to http://localhost:8000/ select "commentator_adk"
agent and click on the "Use camera" button.

To restart a session click on the "New session" button and reload the page.
"""

import os

from genai_processors.core import adk
import commentator


# You need to define the API key in the environment variables.
# export GOOGLE_API_KEY=...
API_KEY = os.environ['GOOGLE_API_KEY']


root_agent = adk.ProcessorAgent(
    (lambda: commentator.create_live_commentator(API_KEY)),
    name='commentator_adk',
)



# %%
# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Wraps a Hugging Face Transformers model into a Processor.

This module allows running Processor pipelines with locally-run LLMs, such as
Gemma. Also before working with Gemma models, make sure you have requested
access via Kaggle [https://ai.google.dev/gemma/docs/setup#get-access] and
reviewed the Gemma terms of use [https://ai.google.dev/gemma/terms].
"""

import asyncio
from collections.abc import AsyncIterable
import dataclasses
import json
import re
from typing import Any, Callable, Literal

from absl import logging
from genai_processors import content_api
from genai_processors import processor
from genai_processors import tool_utils
from google.genai import types as genai_types
import transformers
from typing_extensions import TypedDict


class GenerateContentConfig(TypedDict, total=False):
  """Optional model configuration parameters."""

  system_instruction: content_api.ProcessorContentTypes
  """Instructions for the model to steer it toward better performance.

  For example, "Answer as concisely as possible" or "Don't use technical
  terms in your response".
  """

  seed: int | None
  """Seed."""

  stop_sequences: list[str]
  """Stop sequences."""

  temperature: float | None
  """Controls the randomness of predictions."""

  top_k: float | None
  """If specified, top-k sampling will be used."""

  top_p: float | None
  """If specified, nucleus sampling will be used."""

  tools: list[genai_types.Tool | Callable[..., Any]] | None
  """Tools the model may call."""

  max_output_tokens: int | None
  """Maximum number of tokens that can be generated in the response."""


class TransformersModel(processor.Processor):
  """`Processor` that calls the Hugging Face Transformers model.

  Note: All content is buffered prior to calling the model.

  IMPORTANT: The model does NOT maintain state. You must pass the full
  conversation history every time if you want the model to be aware of previous
  turns.
  """

  def __init__(
      self,
      *,
      model_name: str = '',
      generate_content_config: GenerateContentConfig | None = None,
      log_chat_template: bool = False,
      tool_response_format: Literal['dict', 'string'] = 'string',
  ):
    """Initializes the Transformers model.

    Args:
      model_name: Pretrained model name or path.
      generate_content_config: Inference settings.
      log_chat_template: Whether to log the HF chat template. This is useful for
        debugging, but can be verbose.
      tool_response_format: The format of the tool response. By default, the tool response is returned as a
        string.

    Returns:
      A `Processor` that calls Hugging Face Transformers model in turn-based
      fashion.
    """  # fmt: skip
    self._log_chat_template = log_chat_template
    self._tool_response_format = tool_response_format
    self._generate_content_config = generate_content_config or {}

    self._hf_processor = transformers.AutoProcessor.from_pretrained(model_name)
    self._model = transformers.AutoModelForCausalLM.from_pretrained(
        model_name, device_map='auto'
    )
    self._tools = []
    self._parse_function_calls = False
    if tools_config := self._generate_content_config.get('tools'):
      self._parse_function_calls = True
      for fdecl in tool_utils.to_function_declarations(tools_config):
        self._tools.append(tool_utils.function_declaration_to_json(fdecl))

    self._system_instruction = []
    for part in content_api.ProcessorContent(
        self._generate_content_config.get('system_instruction', ())
    ):
      self._system_instruction.append(
          _to_hf_message(
              part,
              tool_response_format=self._tool_response_format,
              default_role='system',
          )
      )
    self._generation_kwargs = {}
    for arg in ['temperature', 'top_k', 'top_p']:
      if self._generate_content_config.get(arg) is not None:
        self._generation_kwargs[arg] = self._generate_content_config[arg]

    self._generation_kwargs['max_new_tokens'] = (
        self._generate_content_config.get(
            'max_output_tokens', self._model.config.max_position_embeddings
        )
    )

    if seed := self._generate_content_config.get('seed'):
      transformers.set_seed(seed)

  async def call(
      self, content: AsyncIterable[content_api.ProcessorPartTypes]
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    """Internal method to call the Ollama API and stream results."""
    messages = list(self._system_instruction)
    async for part in content:
      messages.append(
          _to_hf_message(
              part,
              tool_response_format=self._tool_response_format,
              default_role='user',
          )
      )
    if not messages:
      return

    inputs = self._hf_processor.apply_chat_template(
        messages,
        tools=self._tools,
        add_generation_prompt=True,
        return_dict=True,
        return_tensors='pt',
    )

    if self._log_chat_template:
      inputs_str = self._hf_processor.apply_chat_template(
          messages,
          tools=self._tools,
          add_generation_prompt=True,
          return_tensors='pt',
          tokenize=False,
      )
      logging.info('HF CHAT TEMPLATE: %s', inputs_str)

    input_len = len(inputs['input_ids'][0])
    output_queue = asyncio.Queue[content_api.ProcessorPart | None]()

    generate_task = processor.create_task(
        asyncio.to_thread(
            self._model.generate,
            **inputs.to(self._model.device),
            streamer=_Streamer(
                input_len,
                asyncio.get_running_loop(),
                self._hf_processor,
                output_queue,
                parse_function_calls=self._parse_function_calls,
            ),
            pad_token_id=self._hf_processor.eos_token_id,
            **self._generation_kwargs,
        )
    )

    try:
      while (part := await output_queue.get()) is not None:
        yield part
    finally:
      await generate_task


def _to_hf_message(
    part: content_api.ProcessorPart,
    tool_response_format: str,
    default_role: str = '',
) -> dict[str, Any]:
  """Returns HF message JSON."""
  # Gemini API uses upper case for roles, while transformers uses lower case.
  role = part.role.lower() or default_role
  if role == 'model':
    role = 'assistant'

  message: dict[str, Any] = {'role': role}

  if part.function_call:
    message['role'] = 'assistant'
    message['tool_calls'] = [{
        'type': 'function',
        'function': {
            'name': part.function_call.name,
            'arguments': part.function_call.args,
        },
    }]
    return message
  elif part.function_response:
    message['role'] = 'tool'
    message['name'] = part.function_response.name
    response = part.function_response.response
    match tool_response_format:
      case 'string':
        if 'result' in response:
          response = response['result']
        message['content'] = (
            json.dumps(response) if not isinstance(response, str) else response
        )
      case 'dict':
        if 'result' in response and isinstance(response['result'], dict):
          response = response['result']
        message['content'] = {
            'name': part.function_response.name,
            'response': response,
        }
      case _:
        raise ValueError(
            f'Unsupported tool response format: {tool_response_format}'
        )
    return message
  elif content_api.is_text(part.mimetype):
    message['content'] = [{'type': 'text', 'text': part.text}]
  elif content_api.is_image(part.mimetype):
    raise ValueError('Images are not supported yet.')
    # TODO(kibergus): Add image support. Can they be passed as data and not URL?
    # message['content'] = [{
    #     "type": "image",
    #     "image": [{"type": "image", "url": part.text},
    # ]
  else:
    raise ValueError(f'Unsupported Part type: {part.mimetype}')

  return message


@dataclasses.dataclass(frozen=True)
class _Tokens:
  segment: Literal['function_call', 'text']
  ids: list[int]


class _Streamer(transformers.generation.BaseStreamer):
  """A utility class for streaming tokens out of a transformer models.

  As the model generation is run in a separate thread we use asyncio.Queue and
  asyncio.Loop.call_soon_threadsafe to push the output back to the event loop.
  """

  def __init__(
      self,
      skip_tokens: int,
      loop: asyncio.AbstractEventLoop,
      hf_processor: transformers.AutoProcessor,
      output_queue: asyncio.Queue[content_api.ProcessorPartTypes | None],
      parse_function_calls: bool = True,
  ):
    self._skip_tokens = skip_tokens
    self._loop = loop
    self._hf_processor = hf_processor
    self._queue = output_queue
    self._parse_function_calls = parse_function_calls
    # To mitigate prompt injection attacks we parse function calls directly from
    # tokens and not from text representation. This assumes that the model is
    # trained to use dedicated tokens for the start and end of the call and
    # prompt tokenization correctly escapes attempts to pass fake function
    # calls.
    self._current_function_call_tokens = None
    # Collect the token ids for the start and end of the function call. They
    # will be used to detect the tokens corresponding to a full function call.
    function_call_token_ids = hf_processor.encode(
        text='<start_function_call><end_function_call><escape>',
        add_special_tokens=False,
    )
    assert len(function_call_token_ids) == 3
    self._start_function_id = function_call_token_ids[0]
    self._end_function_id = function_call_token_ids[1]
    self._escape_id = function_call_token_ids[2]
    # Pattern to capture the function name and arguments as groups.
    self._function_call_pattern = re.compile(
        r'^<start_function_call>call:([^{]+)(\{.*\})<end_function_call>$'
    )

  def _process_function_call_tokens(self, tokens: list[int]) -> list[_Tokens]:
    """Extract function calls from the tokens.

    The tokens list can include many function calls in a row (potentially). This
    method will extract the function calls and return the tokens before and
    after each function call. The returned list contains _Tokens segments
    that are either text or a complete function call.

    When _Tokens.segment == 'function_call', the tokens include the function
    call segment corresponding to:
    <start_function_call>call:fn_name...<end_function_call>.

    Args:
      tokens: The token ids list to extract function calls from.

    Returns:
      A list of _Tokens segments.
    """
    if not tokens:
      return []
    if not self._parse_function_calls:
      return [_Tokens(segment='text', ids=tokens)]
    if self._current_function_call_tokens is None:
      # We are not in a function call.
      try:
        lind = tokens.index(self._start_function_id)
      except ValueError:
        # No function call start token id found, return the tokens as is.
        return [_Tokens(segment='text', ids=tokens)]
      # We have found a function call start token id, start collecting the
      # tokens that are part of the function call.
      self._current_function_call_tokens = []
      remaining_tokens = self._process_function_call_tokens(tokens[lind:])
      return [
          _Tokens(segment='text', ids=tokens[:lind]),
          *remaining_tokens,
      ]
    else:
      try:
        rind = tokens.index(self._end_function_id)
      except ValueError:
        # No function call end token id found, collect the tokens for the
        # current function call.
        self._current_function_call_tokens.extend(tokens)
        return []
      # We have found the function call end token id, yield the tokens
      # collected so far and process the remaining tokens.
      fc_tokens = self._current_function_call_tokens + tokens[: rind + 1]
      self._current_function_call_tokens = None
      remaining_tokens = self._process_function_call_tokens(tokens[rind + 1 :])
      return [
          _Tokens(segment='function_call', ids=fc_tokens),
          *remaining_tokens,
      ]

  def _extract_function_call_part(
      self, part: str
  ) -> content_api.ProcessorPartTypes:
    """Extracts a structured function call from its string representation."""
    fc_match = self._function_call_pattern.match(part)
    if not fc_match:
      return part
    fc_name = fc_match.group(1)
    try:
      fc_args = json.loads(fc_match.group(2))
    except json.JSONDecodeError as e:
      return content_api.ProcessorPart(
          f'Could not parse function call arguments: {fc_match.group(2)} for'
          f' function call {fc_name}: {e}',
          substream_name=processor.DEBUG_STREAM,
          role='model',
      )

    return content_api.ProcessorPart.from_function_call(
        name=fc_name,
        args=fc_args,
        role='model',
    )

  def _add_quotes_on_properties(self, token_str: str) -> str:
    """Adds quotes around the property name in a function call."""
    return re.sub(r'([{,]\s*)(\w+)\b:', r'\1"\2":', token_str)

  def put(self, value):
    """Invoked to push new tokens."""
    tokens = value.flatten()
    tokens_to_skip = min(len(tokens), self._skip_tokens)
    tokens = tokens[tokens_to_skip:]
    self._skip_tokens -= tokens_to_skip
    # tokens are torch tensors, we convert them to list here.
    for tokens in self._process_function_call_tokens(tokens.cpu().tolist()):
      if tokens.segment == 'function_call':
        # We have found a function call, process the escape tokens.
        text_tokens = []
        token_ids = []
        inside_escape = False
        for token in tokens.ids:
          # Function arguments may contain special symbols like " and \. While
          # models can escape them, sometimes they get confused, especially if
          # the tool parameter is a code and contains escaping on its own.
          # By using a special <escape> token to separate parameters we can
          # achieve a more robust operation: we escape all the text in between
          # <escape> tokens and parse the result as JSON.
          if token == self._escape_id:
            token_str = self._hf_processor.decode(
                token_ids, skip_special_tokens=True
            )
            if inside_escape:
              # We are inside a string argument, keep things as is but escape
              # all quotes inside to be parsed correctly as JSON.
              text_tokens.append(json.dumps(token_str))
            else:
              # We are not inside a string argument, the property name should be
              # quoted to be parsed correctly as JSON.
              text_tokens.append(self._add_quotes_on_properties(token_str))
            token_ids = []
            inside_escape = not inside_escape
          else:
            token_ids.append(token)
        if token_ids:
          text_tokens.append(
              self._add_quotes_on_properties(
                  self._hf_processor.decode(token_ids, skip_special_tokens=True)
              )
          )
        part = ''.join(text_tokens)
        # Parse the function call arguments as plain json object.
        part = self._extract_function_call_part(part)
      else:
        part = self._hf_processor.decode(tokens.ids, skip_special_tokens=True)
      if part:
        self._loop.call_soon_threadsafe(self._queue.put_nowait, part)

  def end(self):
    """Invoked to signal the end of generation."""
    self._loop.call_soon_threadsafe(self._queue.put_nowait, None)



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for managing part streams."""

import asyncio
from collections.abc import AsyncIterable, AsyncIterator, Iterable
import copy
from typing import Any, TypeVar, Union, overload

from genai_processors import context

_T = TypeVar('_T')


def split(
    content: AsyncIterable[_T],
    *,
    n: int = 2,
    with_copy: bool = False,
) -> tuple[AsyncIterable[_T], ...]:
  """Split a stream into `n` identical streams.

  Recommended to be used with processor.context to ensure error propagation.

  Args:
    content: content to be split
    n: number of streams to return
    with_copy: whether to copy the items of the streams or not. It is
      recommended to copy the items when side effects between streams can
      happen. This is the case when one processor changes a part in place (e.g.
      update its metadata). As this can be expensive if the items are large and
      the number of streams is high, the default is to not copy. Consider
      setting this to True if there is a chance that a part can be modified in
      place.

  Returns:
    n streams of content.

  Raises:
    ValueError if n=0
  """
  if n == 0:
    raise ValueError('Cannot split a stream in n=0 streams.')
  if n == 1:
    return (content,)
  queues = [asyncio.Queue() for _ in range(n)]

  async def enqueue_parts() -> None:
    async for part in content:
      for queue in queues:
        if with_copy:
          queue.put_nowait(copy.deepcopy(part))
        else:
          queue.put_nowait(part)
    for queue in queues:
      queue.put_nowait(None)

  async def dequeue_parts(
      queue: asyncio.Queue[_T],
  ) -> AsyncIterable[_T]:
    while (part := await queue.get()) is not None:
      yield part

  context.create_task(enqueue_parts())

  return tuple(dequeue_parts(queue) for queue in queues)


async def concat(*contents: AsyncIterable[_T]) -> AsyncIterable[_T]:
  """Concatenate multiple streams into one.

  The streams are looped over concurrently before being assembled into a single
  output stream.

  Args:
    *contents: each stream to concat as a separate argument.

  Yields:
    The concatenation of all streams.
  """
  output_queues = [asyncio.Queue() for _ in contents]

  async def _stream_outputs(
      idx: int,
  ):
    async for c in contents[idx]:
      output_queues[idx].put_nowait(c)
    # Adds None to indicate end of output.
    output_queues[idx].put_nowait(None)

  tasks = []
  for idx, _ in enumerate(contents):
    tasks.append(context.create_task(_stream_outputs(idx)))

  for q in output_queues:
    while (part := await q.get()) is not None:
      q.task_done()
      yield part


# 1. Overload for the *args style
@overload
def merge(
    *streams: AsyncIterable[_T],
    queue_maxsize: int = 0,
    stop_on_first: bool = False,
) -> AsyncIterable[_T]:
  ...


# 2. Overload for the single Iterable (list/tuple) style
# Deprecated and is provided for backward compatibility.
# Please prefer the merge(stream1, stream2, ...) syntax.
@overload
def merge(
    streams: Iterable[AsyncIterable[_T]],
    *,
    queue_maxsize: int = 0,
    stop_on_first: bool = False,
) -> AsyncIterable[_T]:
  ...


async def merge(
    *args: Union[Iterable[AsyncIterable[_T]], AsyncIterable[_T]],
    queue_maxsize: int = 0,
    stop_on_first: bool = False,
) -> AsyncIterator[_T]:
  """Merges multiple streams into one.

  The order is defined by the asyncio loop and will likely be determined by the
  time when the items are available.

  If a stream is cancelled, the overall merge will be cancelled and all other
  streams will be cancelled as well.

  Args:
    *args: The input streams to merge, either as a single iterable of
      AsyncIterables or as multiple AsyncIterable positional arguments. These
      streams cannot be iterated over outside of this call. If you need to
      consume them outside of this call, use `streams.split()` to copy the
      stream first.
    queue_maxsize: The maximum number of items to buffer in an internal queue
      for each input stream. Set to 0 to use an unbounded queue.
    stop_on_first: If True, stop merging streams as soon as one of them is
      empty.

  Yields:
    An item from one of the input streams. The order in which the items are
    yielded is random and interleaved (likely order of generation). This means
    the order of the items within one stream is preserved but not across
    streams.
  """

  # Determine if the first arg is the collection or a single stream
  if len(args) == 1 and not hasattr(args[0], '__aiter__'):
    streams = list(args[0])
  else:
    streams = list(args)

  if not streams:
    return

  out = asyncio.Queue(maxsize=queue_maxsize)
  active_streams = 0

  async with asyncio.TaskGroup() as tg:
    running_tasks = []
    for s in streams:
      active_streams += 1
      running_tasks.append(tg.create_task(enqueue(s, out)))

    while active_streams:
      part = await out.get()
      out.task_done()
      if part is None:
        active_streams -= 1
        if stop_on_first:
          break
        continue
      yield part
    for t in running_tasks:
      t.cancel()


async def enqueue(
    content: AsyncIterable[_T], queue: asyncio.Queue[_T | None]
) -> None:
  """Enqueues all content into a queue.

  When the queue is unbounded, this function will not block. When the queue is
  bounded, this function will block until the queue has space.

  Args:
    content: The content to enqueue.
    queue: The queue to enqueue to.
  """
  try:
    async for part in content:
      await queue.put(part)
  finally:
    await queue.put(None)


async def dequeue(queue: asyncio.Queue[_T | None]) -> AsyncIterator[_T]:
  """Dequeues content from a queue.

  Args:
    queue: The queue to dequeue from. The queue must end with a None item.

  Yields:
    The content from the queue as an AsyncIterator. The items are yielded in
    the order they were enqueued.
  """
  while (part := await queue.get()) is not None:
    queue.task_done()
    yield part
  queue.task_done()


async def stream_content(
    content: Iterable[_T],
    with_delay_sec: float | None = None,
    delay_first: bool = False,
    delay_end: bool = True,
) -> AsyncIterable[_T]:
  """Converts non-async content into an AsyncIterable.

  Args:
    content: the items to yield. The state of the iterator provided by `content`
      is undefined post-invocation and should not be relied on.
    with_delay_sec: If set, asyncio.sleep() is called with this value between
      yielding parts, and optionally also before and after.
    delay_first: If set to True, a delay is added before the first part.
    delay_end: Unless set to False, a delay is added between the last part and
      stopping the returned AsyncIterator.

  Yields:
    an item in `content` (unchanged).
  """
  # Getting the next value from content must not block. For this reason
  # this function doesn't work for generators.
  delay_next = with_delay_sec is not None and delay_first
  for c in content:
    if delay_next:
      await asyncio.sleep(with_delay_sec)
    else:
      delay_next = with_delay_sec is not None  # From the 2nd iteration onwards.
    yield c
  if with_delay_sec is not None and delay_end:
    await asyncio.sleep(with_delay_sec)


async def gather_stream(content: AsyncIterable[_T]) -> list[_T]:
  """Gathers an AsyncIterable into a list of items."""
  return [c async for c in content]


async def aenumerate(
    aiterable: AsyncIterable[_T],
) -> AsyncIterable[tuple[int, _T]]:
  """Enumerate an async iterable."""
  i = 0
  async for x in aiterable:
    yield (i, x)
    i += 1


async def endless_stream() -> AsyncIterable[Any]:
  """Empty input stream for the live agents.

  The Live API and VideoIn / PyAudioIn processors' lifetime is bound to the
  incoming Part stream. More complex setups would use the stream to send
  additional data such as the initial context or text entered in the terminal.

  For simpler use cases this is an empty never-ending stream to keep such
  processors alive.

  Yields:
    Nothing. The stream keeps waiting and will never end unless cancelled.
  """
  while True:
    await asyncio.sleep(1)
  # Unreachable. Needed to make the function a generator.
  yield None



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

r"""Live commentator using GenAI Processors.

You can run this example from a CLI directly or from AI Studio.

See commentator_cli.py or commentator_ais.py for usage.

The commentator observes the incoming video/audio feed and produces an audio
commentary on it. While it is the agent who drives the conversation, it is
interactive and can be interruped/corrected by a user.

This example makes good use of the various ways the Gemini API can receive
content:
 * `system_instruction`: guidelines for the model provided by the agent
   developers. They are hardcoded in this file in the PROMPT_PARTS variable.
 * `session.send_client_content` is used for turn-by-turn scenarios and is
   analogous to what `client.generate_content` does. We use it to provide any
   additional context (non audio and video feeds) to the model. The
   `event_detection` processor uses this option as it is optimal for its fixed
   cadence, turn-based operation.
 * `session.send_realtime_input` is used for realtime interactive input and has
   built-in voice activation detection (VAD). This allows us to have lower
   latency and makes the model automatically respond to user's utterances. Note
   that audio commentary from the model itself needs to be filtered out to avoid
   VAD activating erroneously. We rely on the echo cancellation built into the
   browser for that.
 * **Async function calls**: Gemini API now supports function calls that don't
   block the generation and can even stream their output back to the model. If
   the INTERRUPT scheduling strategy is used, they will interrupt ongoing
   inference and restart it with the latest information available.

The standard Gemini Live API is well suited for the use case where the user
drives the conversation and the model responds when prompted. The live
commentator agent, on the contrary, drives the conversation, keeps commentating
while still reacting to events and to what the user asks.

A naive approach would be to generate one second of commentary every second.
However, models can produce output much faster than realtime, and forcing them
to be in a lockstep with the real world will either leave the TPUs idle or
induce a huge overhead on maintaining or recalculating the KV cache. Also
text-to-speech works better if it can work on longer sentences.

So, we employ a more event-driven approach:

 * Let the model produce long-ish commentary.
 * Slow down the streaming of these comments to real-time using the
   `RateLimitAudio` processor.
 * Rely on Gemini API built-in voice activation detection to notice if the user
   speaks over, corrects, or asks the agent.
 * Run a cheap model to detect whether something notable has happened that the
   agent should interrupt its current speech and generate a new phrase.

The async function calls play the crucial role in driving the interruption
process. We register the `start_commentating` tool which drives the timing when
the model should be making comments. It is the model itself, not a tool, that
produces comments. By using `behavior="NON_BLOCKING"` the tool is able to
schedule the generation of the next comment without interrupting the current
one. It also makes the model ingest the latest observations and tells the model
whether it should continue its previous comment or react to an interruption.

Async function calls also allow the model to send structured signals to the
client while the default output is Audio. We use a `wait_for_user` async
function to let the model decide when to pause the commentary and wait for the
user to do something. As soon as the client receives the `wait_for_user` signal,
it pauses the commentary and sends back an empty response that does not trigger
any model generate call. The silent response is specific to async function calls
and enables this "fire-and-forget" calling style.

By representing the agent as an async function we also give the model ability to
 start and stop commentating at will. If the user asks to stop commentating, the
 model will cancel the tool call and turn into a regular voice activated agent.
 When asked to start commentating, the model will invoke `start_commentating`
 tool, which will awake the event detection loop.
"""

import asyncio
import collections
import dataclasses
import enum
import random
import re
import time
from typing import Any, AsyncIterable, Optional

from absl import logging
from genai_processors import content_api
from genai_processors import processor
from genai_processors import streams
from genai_processors.core import event_detection
from genai_processors.core import genai_model
from genai_processors.core import live_model
from genai_processors.core import rate_limit_audio
from genai_processors.core import text
from genai_processors.core import timestamp
from google.genai import types as genai_types
import numpy as np


# Model to use for the live api processor.
MODEL_LIVE = 'gemini-2.5-flash-native-audio-preview-12-2025'

# Model to use for the event detection processor.
MODEL_DETECTION = 'gemini-2.5-flash-lite'

# Number of times the detection should detect a "no detection" in a row before
# stopping the commentator.
NO_DETECTION_SENSITIVITY = 3

# Audio config
RECEIVE_SAMPLE_RATE = 24000

# Maximum number of seconds the commentator will wait for a user signal before
# continuing the commentary.
MAX_SILENCE_WAIT_FOR_USER_SEC = 5

# Number of seconds to wait before checking if we need to comment again. This is
# used when a comment is blocked because of chattiness.
NO_COMMENT_DELAY_SEC = 3

# User message when the agent detects it should continue commentating.
COMMENT_MSG = (
    'Continue commentating on what you see, do not repeat the same comments.'
    ' Always address the persons in the camera as "you" and assume they can'
    ' hear you and are in the same room. Remember this is a conversation.'
    ' Do not cancel the commentating unless the user explicitly asks you to.'
)
# Response when the agent detects something special that should interrupt the
# commentating while it is ongoing.
INTERRUPT_COMMENT_MSG = (
    'Interrupt the current commentary to comment what you see now. If there is'
    ' no commentary going on, start a new one. Start your comment as if you'
    ' were interrupting yourself or interrupting a silence. Then commentate on'
    ' the new event that just triggered this interruption. If this event is'
    ' related to what you asked the user to do, continue the previous comment'
    ' or instructions. Do not cancel the commentary.'
)

# Response when the agent has waited too long for the user to respond.
WAIT_FOR_USER_MSG = (
    'Resume commentating. Do not repeat the same comments but check if anything'
    ' explains why the user is not responding.'
)

# Response when the agent is waiting for a user signal and receives an interrupt
# from the event detection processor.
INTERRUPT_WAIT_FOR_USER_MSG = (
    'Check if you need to make a `wait_for_user` call to wait for something to'
    ' happen. If you make a `wait_for_user` call, stay silent, do not output'
    ' anything but this call. Otherwise, start commentating, or answer the user'
    ' request or move to the next steps of your instructions or of the'
    ' conversation.'
)

# Response when the agent is saying something wrong or unsafe.
START_AGAIN_MSG = (
    'There was a glitch, I could not get your last message, what were you'
    ' saying?'
)

# Async function declarations.
TOOLS = [
    genai_types.Tool(
        function_declarations=[
            # start_commentating tool drives the timing when the model should
            # be making comments. It is the model itself, not a tool, that
            # produces comments. By using `behavior="NON_BLOCKING"` the tool is
            # able to schedule the generation in many ways (controlled by the
            # `scheduling` field in the function response) and to make the model
            # include the latest observations in its prompt before generation.
            genai_types.FunctionDeclaration(
                name='start_commentating',
                description=(
                    'Starts commentating on the video feed. The model should'
                    ' continue commentating until the user asks to stop. Caller'
                    ' must print() the result for this to work. Do not cancel'
                    ' the commentating unless the user asks you to.'
                ),
                behavior='NON_BLOCKING',
            )
        ]
    ),
    genai_types.Tool(
        function_declarations=[
            genai_types.FunctionDeclaration(
                name='wait_for_user',
                description=(
                    'Waits for the user to respond to your question or to do'
                    ' something special on the video stream, e.g. user picks up'
                    ' a pen & a piece of paper. This function will return'
                    ' quickly with an empty response. The next user request'
                    ' implies there has been a user response. It is up to the'
                    ' caller of this function to decide then whether to wait'
                    ' for the user again.'
                ),
                # We use the non-blocking behavior to ensure the model will not
                # output anything when receiving the function response:
                # scheduling is set to `SILENT`.
                behavior='NON_BLOCKING',
            )
        ]
    ),
]

# Prompt for the live api processor.
PROMPT_PARTS = [
    (
        'You are an agent that commentates on a video feed and'
        ' interacts with the user in a conversation. Commentate'
        ' highlighting the most important things and making it lively'
        ' and interesting for the user. Approach the commentating as if'
        ' you were on TV commentating and interviewing at the same'
        ' time. You can make jokes, explain interesting facts related'
        ' to what you see and hear, predict what could happen, judge'
        ' some actions or reactions, etc.'
    ),
    (
        'You can be interrupted in the middle of a commentary. If this'
        ' happens, you should answer the question from the user first.'
        ' Then you can continue commentating. If the user asks you to'
        ' adjust your style or tone, you should do so.'
    ),
    (
        'You can be asked to do many different things, such as teaching a skill'
        ', helping drawing something, explaing how to go into a yoga position,'
        ' or play a game like Simon says. You should do your best to help and'
        ' drive the conversation in that case, that is, be the coach or teacher'
        ' for the user in the video stream.'
    ),
    (
        'In some situations, you will need to wait for the user to do something'
        ' or to respond to your question or instructions. When this happens, be'
        ' silent and call `wait_for_user` to let the user respond or do what'
        ' they were instructed to do. Do not say `waiting for your answer` to'
        ' the user after or before you call `wait_for_user` and if you asked a'
        ' question.'
    ),
    (
        'When something changes in the image, video or audio, you'
        ' should commentate about it interrupting what you were saying'
        ' before. It is important to stay relevant to what is happening'
        ' recently and not in the images long before.'
    ),
    (
        'Always commentate assuming the persons in the video stream hear you'
        ' and are in the same room as you. You should always address'
        ' the persons in the video stream as in a normal conversation and not'
        ' say he or she when referring to them.'
    ),
    (
        'Do not make long comments at each turn, just one or two sentences. You'
        ' should not repeat the same comments twice.'
    ),
    (
        'Remember to always address the persons in the video stream in the'
        ' first person, and not he or she. You can address them by their'
        ' location or how they are dressed.'
    ),
]

# Higher resolution is needed for better text understanding.
# MEDIA_RESOLUTION_MEDIUM is the default but for higher FPS MEDIA_RESOLUTION_LOW
# might be a better fit to reduce latency and conserve tokens.
MEDIA_RESOLUTION = genai_types.MediaResolution.MEDIA_RESOLUTION_MEDIUM


class EventTypes(enum.StrEnum):
  DETECTION = 'yes'
  NO_DETECTION = 'no'
  INTERRUPTION = 'interruption'


EVENT_DETECTION_PROMPT = (
    'You are an agent that detects events in the video feed. You receive images'
    ' from a camera. You must detect whenever a person or a group of people are'
    ' facing the camera and are close enough to engage in a conversation. You'
    ' must as well detect whenever the camera is pointing at a computer'
    ' screen.\n'  # Do not fold
    'Whenever you see a person or a group of people close enough to the camera,'
    f" respond with '{EventTypes.DETECTION}'.\n"  # Do not fold
    f"Whenever you see a computer screen, respond '{EventTypes.DETECTION}'.\n"
    'Whenever you see something special, the user wanting to stop you, interact'
    ' with you or whenever you see something new that would change your'
    f" comment, respond with '{EventTypes.INTERRUPTION}'. Do not interrupt if"
    ' the user is talking to you, only when something related to the video feed'
    'changes significantly.\n'
    f"In all other cases, respond with '{EventTypes.NO_DETECTION}'.\n"
)


def audio_duration_sec(audio_data: bytes, sample_rate: int) -> float:
  """Returns the duration of the audio data in seconds."""
  # 2 bytes per sample (16bits), 24kHz sample rate
  return len(audio_data) / (2 * sample_rate)


class GenerationType(enum.Enum):
  """Type of the request that triggers a model generation call."""

  # Comment scheduled by the commentator.
  COMMENT = 1
  # Question from the user.
  USER_REQUEST = 2
  # Special event detected during the conversation.
  EVENT_INTERRUPTION = 3


@dataclasses.dataclass
class GenerationRequestInfo:
  """Info on a model Generate call request.

  We have three types of request:

  - USER_REQUEST: user is talking.
  - COMMENT: scheduled comment.
  - EVENT_INTERRUPTION: special event detected during the conversation.

  A generation request is active from the moment we sent a request to the model
  until the model returns the `generation_complete` signal. This means we have
  received all the audio parts we need. When we're after the
  `generation_complete` signal, the request is considered inactive. This is in
  contrast with the server side where the request will stay active until the
  model returns the `turn_complete` or `interrupt` signal.
  """

  # Time when the model call was made.
  generation_start_sec: Optional[float] = None
  # Type of the generation request.
  generation_type: Optional[GenerationType] = None
  # Time when the audio response stream started. None if no audio stream was
  # received.
  time_audio_start: Optional[float] = None
  # TTFT of the model call. None if no model call was made or if a model call is
  # in progress.
  ttft_sec: Optional[float] = None
  # Time of the audio response stream received so far.
  audio_duration: float = 0.0

  def update(self, media_blob: genai_types.Blob):
    """Updates the generation request with the new media data.

    Args:
      media_blob: The new media data.
    """
    if self.generation_start_sec is not None and self.ttft_sec is None:
      # New comment is starting to be streamed.
      self.time_audio_start = time.perf_counter()
      self.ttft_sec = self.time_audio_start - self.generation_start_sec
    self.audio_duration += audio_duration_sec(
        media_blob.data,
        RECEIVE_SAMPLE_RATE,
    )


class State(enum.Enum):
  """Finite set of states that the commentator can be in."""

  # Commentator is not commentating. User can still ask questions but there is
  # no comment request scheduled.
  OFF = 1
  # Commentator is commentating or responding to the user.
  TALKING = 2
  # User starts talking or, more precisely, the model has detected that the
  # audio-input is a user barging in.
  USER_IS_TALKING = 3
  # An event is detected in the video stream. This state is reached when the
  # commentator receives an event interruption. It ends when the model returns
  # the first audio part of the interruption comment.
  REQUESTING_INTERRUPTION = 4
  # The commentator is requesting a new comment. This state is reached when the
  # commentator sends a comment request to the model. It ends when the model
  # returns the first audio part of the comment.
  REQUESTING_COMMENT = 5
  # The commentator is requesting a response to the user. This state is
  # reached when the commentator sends a user request to the model. It ends
  # when the model returns the first audio part of the response.
  REQUESTING_RESPONSE = 6
  # The commentator is interrupting from a detection but has not received audio
  # parts yet.
  INTERRUPTED_FROM_DETECTION = 7
  # After talking is done, the commentator will wait for the user to respond to
  # a question or instruction, or to an interruption to occur.
  WAITING_FOR_USER = 8


class Action(enum.Enum):
  """Action that the commentator can take."""

  # Turns on the commentator.
  TURN_ON = 1
  # Turns off the commentator.
  TURN_OFF = 2
  # Starts streaming the media part back to the user. This is the first audio
  # part of a comment, an interruption or a response to a user request.
  STREAM_MEDIA_PART = 3
  # Sends a model call to respond to the user (text based).
  REQUEST_FROM_USER = 4
  # Sends a model call to react to an interruption from the event detection.
  REQUEST_INTERRUPT = 5
  # Sends a model call to get the next comment.
  REQUEST_FROM_COMMENTATOR = 6
  # Interrupts the current request (user barging in or event interruption).
  INTERRUPT = 7
  # Wait for the user to say or do something.
  WAIT_FOR_USER = 8


@dataclasses.dataclass
class CommentatorStateMachine:
  """(state, action) -> state transitions for the commentator."""

  # State of the commentator.
  state: State = State.OFF
  # Generation requests currently in progress, aka active.
  generation_request_info: Optional[GenerationRequestInfo] = None
  ttfts: collections.deque[float] = dataclasses.field(
      default_factory=collections.deque
  )
  # Commentator ID, defined by the Async Fn call that triggered the commentator.
  id: Optional[str] = None

  def update(self, action: Action, state_arg: Any = None) -> None:
    """Updates the commentator state for the new action."""
    start_state = self.state
    try:
      match (self.state, action):
        # --- Turn On/Off ---
        case (State.OFF, Action.TURN_ON | Action.INTERRUPT):
          if isinstance(state_arg, str):
            self.id = state_arg
          # When commentator is turned on, the model outputs a comment.
          self.state = State.TALKING
          # Turn on can only happen if a generation happened from the user.
          self.update(Action.REQUEST_FROM_COMMENTATOR)
        case (State.OFF, _):
          # Do nothing if the commentator is off and it's not a TURN_ON action.
          pass
        case (_, Action.TURN_OFF):
          self.state = State.OFF
          self.generation_request_info = None
          self.id = None
        # --- Interruptions from Event Detection ---
        case (
            State.TALKING | State.WAITING_FOR_USER | State.REQUESTING_COMMENT,
            Action.REQUEST_INTERRUPT,
        ):
          self.mark_start_generation(GenerationType.EVENT_INTERRUPTION)
          self.state = State.REQUESTING_INTERRUPTION
        case (State.REQUESTING_INTERRUPTION, Action.INTERRUPT):
          self.state = State.INTERRUPTED_FROM_DETECTION
        # --- Requests from User ---
        case (_, Action.INTERRUPT):
          self.state = State.USER_IS_TALKING
          self.mark_start_generation(GenerationType.USER_REQUEST)
          # Delay the generation start to model how long the user talked.
          # The interrupt signal is received when user starts talking.
          # `generation_start_sec` should be the time when the user stops
          # talking.
          self.generation_request_info.generation_start_sec = (
              time.perf_counter() + 2
          )
        case (_, Action.REQUEST_FROM_USER):
          self.state = State.REQUESTING_RESPONSE
          self.mark_start_generation(GenerationType.USER_REQUEST)
        # --- Requests from Commentator ---
        case (
            State.TALKING | State.WAITING_FOR_USER,
            Action.REQUEST_FROM_COMMENTATOR,
        ):
          self.state = State.REQUESTING_COMMENT
          self.mark_start_generation(GenerationType.COMMENT)
        case (_, Action.WAIT_FOR_USER):
          # Wait for user can only happen from the model.
          self.state = State.WAITING_FOR_USER
        # --- Streaming Media ---
        case (_, Action.STREAM_MEDIA_PART):
          if isinstance(state_arg, genai_types.Blob):
            self._update_media_blob(state_arg)
          if self.state != State.WAITING_FOR_USER:
            self.state = State.TALKING
    finally:
      logging.debug(
          '%s - Update: %s + %s -> %s',
          time.perf_counter(),
          start_state,
          action,
          self.state,
      )

  def mark_start_generation(self, generation_type: GenerationType):
    logging.debug(
        '%s - Start generation: %s', time.perf_counter(), generation_type
    )
    self.generation_request_info = GenerationRequestInfo(
        generation_start_sec=time.perf_counter(),
        generation_type=generation_type,
    )

  def _update_media_blob(self, media_part: genai_types.Blob):
    """Updates the generation request with the new media data."""
    if not self.generation_request_info:
      logging.debug(
          '%s - No generation request to update.', time.perf_counter()
      )
      return
    if self.generation_request_info.ttft_sec is None:
      self.generation_request_info.update(media_part)
      if self.generation_request_info.ttft_sec is not None:
        self.ttfts.append(self.generation_request_info.ttft_sec)
    else:
      self.generation_request_info.update(media_part)

  def predict_next_ttft(self) -> float:
    """Predict the next TTFT from the history of TTFTs."""
    if not self.ttfts:
      return 0.0
    avg = np.mean(self.ttfts)
    std = np.std(self.ttfts)
    # Subtract the standard deviation to get a lower bound on the next TTFT.
    # Underestimating the TTFT will result in the comment being triggered too
    # late, adding a delay to the conversation but making the comment more
    # aligned (time-wise) with the video stream.
    return max(0.4, avg - std)

  def tentative_trigger_time(self) -> Optional[float]:
    """Returns the tentative time when the commentator will trigger."""
    logging.debug(
        '%s - generation_request_info: %s ttft: %s',
        time.perf_counter(),
        self.generation_request_info,
        self.ttfts,
    )
    if (
        self.state != State.OFF
        and self.generation_request_info
        and self.generation_request_info.time_audio_start is not None
    ):
      return (
          self.generation_request_info.time_audio_start
          # In WAITING_FOR_USER state, the audio duration can be 0. We add a
          # minimum duration to make sure we trigger another model generate
          # at a reasonable time.
          + max(5.0, self.generation_request_info.audio_duration)
          - self.predict_next_ttft()
      )


class LiveCommentator(processor.Processor):
  """Processor generating live commentaries on a video and audio stream..

  The audio and video parts to commentate on should have the `realtime`
  substream name. Any other parts will be considered as parts of a user request.
  """

  def __init__(
      self,
      live_api_processor: live_model.LiveProcessor,
      chattiness: float = 1.0,
      unsafe_string_list: list[str] | None = None,
  ):
    """Initializes the processor.

    Args:
      live_api_processor: The live API processor to use.
      chattiness: Probability of triggering a comment when the model has
        finished talking or every 3 seconds. Set to 0 to disable commenting.
      unsafe_string_list: The strings to use for unsafe content. If None, the
        processor will not block unsafe content. If set, the processor will
        interrupt itself when it sees the string in the output.
    """
    self._processor = live_api_processor
    self._chattiness = chattiness
    self._commentator = CommentatorStateMachine()
    # Historic time to first token (TTFT) of the recent requests.
    # We use it request the next comment just before the current one finishes.
    self.ttfts = collections.deque(maxlen=50)
    self._unsafe_string_list = unsafe_string_list
    if unsafe_string_list is not None:
      pattern = '|'.join(re.escape(s) for s in unsafe_string_list)
      self._processor += text.MatchProcessor(
          pattern=pattern,
          substream_input='output_transcription',
          substream_output='unsafe_regex',
          remove_from_input_stream=False,
          flush_fn=lambda part: part.get_metadata('generation_complete', False)
          or part.get_metadata('interrupted')
          or part.get_metadata('interrupt_request')
          or part.get_metadata('turn_complete')
          or part.get_metadata('go_away'),
      )

  def set_chattiness(self, chattiness: float):
    self._chattiness = chattiness

  def _start_commentating(
      self,
      input_queue: asyncio.Queue[content_api.ProcessorPart],
      message: str = COMMENT_MSG,
      will_continue: bool = False,
      scheduling: genai_types.FunctionResponseScheduling = genai_types.FunctionResponseScheduling.WHEN_IDLE,
  ) -> None:
    """Triggers a comment from the model. Input queue is fed to the model."""
    if self._commentator.id is None:
      logging.debug(
          '%s - No commentator id, ignoring start_commentating: %s',
          time.perf_counter(),
          self._commentator,
      )
    else:
      logging.debug(
          '%s - Triggering start_commentating: %s',
          time.perf_counter(),
          self._commentator.id,
      )
      input_queue.put_nowait(
          content_api.ProcessorPart.from_function_response(
              function_call_id=self._commentator.id,
              name='start_commentating',
              response={'output': message},
              will_continue=will_continue,
              scheduling=scheduling,
          )
      )

  def _stop_commentating(
      self, input_queue: asyncio.Queue[content_api.ProcessorPart], fn_id: str
  ):
    """Cancels a comment from the model. Input queue is fed to the model."""
    input_queue.put_nowait(
        content_api.ProcessorPart.from_function_response(
            function_call_id=fn_id,
            name='start_commentating',
            response={},
            will_continue=False,
            scheduling=genai_types.FunctionResponseScheduling.SILENT,
        )
    )

  def _respond_to_wait_for_user(
      self, input_queue: asyncio.Queue[content_api.ProcessorPart], fn_id: str
  ):
    """Cancels wait_for_user from the model. Input queue is fed to the model."""
    input_queue.put_nowait(
        content_api.ProcessorPart.from_function_response(
            function_call_id=fn_id,
            name='wait_for_user',
            response={},
            will_continue=False,
            scheduling=genai_types.FunctionResponseScheduling.SILENT,
        )
    )

  async def _schedule_comment(
      self,
      at_time: float,
      input_queue: asyncio.Queue[content_api.ProcessorPart],
      message: str = COMMENT_MSG,
  ):
    """Schedules and triggers a comment from the model."""
    if self._chattiness < 1e-6:
      return
    # Wait for the last moment to trigger the comment. This minimizes the
    # delay between any event on the video stream and the comment.
    await asyncio.sleep(max(0, at_time - time.perf_counter()))
    while True:
      chattiness_dice = random.uniform(0, 1)
      if chattiness_dice < self._chattiness:
        logging.debug(
            '%s - Triggering comment: %s ', time.perf_counter(), message
        )
        self._commentator.update(Action.REQUEST_FROM_COMMENTATOR)
        self._start_commentating(
            input_queue, message=message, will_continue=True
        )
        break
      else:
        await asyncio.sleep(NO_COMMENT_DELAY_SEC)

  async def call(
      self, content: AsyncIterable[content_api.ProcessorPart]
  ) -> AsyncIterable[content_api.ProcessorPart]:
    """Run the main conversation loop."""
    schedule_task = None

    def reset_schedule_task():
      if schedule_task is not None and not schedule_task.done():
        schedule_task.cancel()

    # Input queue for the live api processor - we will inject here what we
    # want to send to the model besides the main content stream.
    input_queue = asyncio.Queue()
    input_stream = streams.merge(
        [content, streams.dequeue(input_queue)], stop_on_first=True
    )

    start_time = time.perf_counter()
    async for part in self._processor(input_stream):
      logging.log_every_n_seconds(
          logging.INFO,
          '%s - commentator running for: %s',
          10,
          time.perf_counter(),
          timestamp.to_timestamp(time.perf_counter() - start_time),
      )

      # Handle unsafe content.
      if part.substream_name == 'unsafe_regex':
        logging.info(
            '%s - Unsafe content detected: %s',
            time.perf_counter(),
            part,
        )
        self._commentator.update(Action.REQUEST_FROM_USER)
        input_queue.put_nowait(
            content_api.ProcessorPart(
                START_AGAIN_MSG
                + (
                    ' Do not mention the following expressions in your'
                    " next response: '%s'"
                    % "', '".join(self._unsafe_string_list)
                ),
                role='USER',
                substream_name='realtime',
            )
        )
        continue

      # Handle function calls.
      if part.function_call:
        logging.info(
            '%s - Received tool call: %s',
            time.perf_counter(),
            part,
        )
        fn_id = part.get_metadata('id')
        if part.part.function_call.name == 'start_commentating':
          if self._commentator.state != State.OFF:
            # We already have a comment in progress, ignore this one.
            logging.info(
                '%s - Ignoring start_commentating: %s',
                time.perf_counter(),
                fn_id,
            )
            self._stop_commentating(input_queue, fn_id)
          else:
            self._commentator.update(Action.TURN_ON, fn_id)
        elif part.part.function_call.name == 'wait_for_user':
          if self._commentator.state != State.OFF:
            logging.debug(
                '%s - Received wait_for_user: %s',
                time.perf_counter(),
                fn_id,
            )
            self._respond_to_wait_for_user(input_queue, fn_id)
            reset_schedule_task()
            self._commentator.update(Action.WAIT_FOR_USER, fn_id)
            # Schedule the resume comment now. We will cancel it and reschedule
            # it again if we receive media parts.
            tentative_trigger_time = self._commentator.tentative_trigger_time()
            if tentative_trigger_time is None:
              logging.debug(
                  '%s - No tentative trigger time, using current time',
                  time.perf_counter(),
              )
              tentative_trigger_time = time.perf_counter()
            schedule_task = processor.create_task(
                self._schedule_comment(
                    at_time=tentative_trigger_time
                    + MAX_SILENCE_WAIT_FOR_USER_SEC,
                    input_queue=input_queue,
                    message=INTERRUPT_WAIT_FOR_USER_MSG,
                )
            )
        continue

      # Handle start of turn, considered as a user request.
      if part.get_metadata('start_of_user_turn'):
        self._commentator.update(Action.REQUEST_FROM_USER)
        continue

      # Handle function cancellation:
      if part.tool_cancellation:
        if part.tool_cancellation == self._commentator.id:
          logging.debug(
              '%s - Cancelling comment function call: %s',
              time.perf_counter(),
              self._commentator.id,
          )
          self._commentator.update(Action.TURN_OFF)
          reset_schedule_task()
        continue

      # Handle when the model is done generating. All audios parts have
      # arrived at this point but they have not been all played back yet.
      if part.get_metadata('generation_complete'):
        logging.debug('%s - generation_complete', time.perf_counter())
        yield processor.ProcessorPart(
            '',
            role='model',
            metadata={'generation_complete': True},
        )
        if self._commentator.state != State.OFF:
          # We have received all audio parts, we reschedule the comment.
          reset_schedule_task()
          tentative_trigger_time = self._commentator.tentative_trigger_time()
          if self._commentator.state != State.WAITING_FOR_USER:
            # Schedule the next commentator turn.
            schedule_task = processor.create_task(
                self._schedule_comment(
                    at_time=tentative_trigger_time,
                    input_queue=input_queue,
                )
            )
          else:
            schedule_task = processor.create_task(
                self._schedule_comment(
                    at_time=tentative_trigger_time
                    + MAX_SILENCE_WAIT_FOR_USER_SEC,
                    input_queue=input_queue,
                    message=INTERRUPT_WAIT_FOR_USER_MSG,
                )
            )

      # Handle interruption from the user.
      if part.get_metadata('interrupted'):
        reset_schedule_task()
        self._commentator.update(Action.INTERRUPT)
        if self._commentator.state == State.USER_IS_TALKING:
          logging.debug('%s - Turn interrupted - user', time.perf_counter())
          yield content_api.ProcessorPart(
              '', role='model', metadata={'interrupted': True}
          )
        else:
          logging.debug(
              '%s - Turn interrupted - detection', time.perf_counter()
          )
        continue

      # Handle interrupt request from the event detection. Do not interrupt
      # yet, wait for the interruption to be confirmed by the model.
      if part.get_metadata('interrupt_request'):
        if self._commentator.state == State.WAITING_FOR_USER:
          response = INTERRUPT_WAIT_FOR_USER_MSG
        else:
          response = INTERRUPT_COMMENT_MSG
        self._commentator.update(Action.REQUEST_INTERRUPT)
        if self._commentator.state == State.REQUESTING_INTERRUPTION:
          self._start_commentating(
              input_queue,
              message=response,
              will_continue=True,
              scheduling=genai_types.FunctionResponseScheduling.INTERRUPT,
          )

      if part.get_metadata('go_away'):
        reset_schedule_task()
        return

      # Handle audio parts received from the live API.
      if part.part.inline_data:
        if self._commentator.state == State.INTERRUPTED_FROM_DETECTION:
          logging.debug(
              '%s - Yield interrupt from interruption, audio should stop now',
              time.perf_counter(),
          )
          yield content_api.ProcessorPart(
              '', role='model', metadata={'interrupted': True}
          )
        self._commentator.update(
            Action.STREAM_MEDIA_PART, part.part.inline_data
        )
      else:
        logging.debug('%s - non media part: %s', time.perf_counter(), part)

      yield part

    reset_schedule_task()


def create_live_commentator(
    api_key: str,
    chattiness: float = 1.0,
    unsafe_string_list: list[str] | None = None,
) -> processor.Processor:
  r"""Creates a live commentator.

  A live commentator processor takes audio and video as input and produces
  live commentaries on the audio and video stream. The commentaries are
  interrupted by events detected on the video stream. Input and video streams
  coming from devices should be passed to the processor with the `realtime`
  substream name.

  Args:
    api_key: The API key to use for the model.
    chattiness: Probability of triggering a comment when the model has finished
      talking or every NO_COMMENT_DELAY_SEC seconds. Set to 0 to disable
      commenting.
    unsafe_string_list: a list of strings that should not be sent back to the
      user. This is a sanity check to make sure the model does not output
      anything that is not allowed. None by default means nothing is blocked.
      When set, the commentator will interrupt itself if the model outputs this
      string and will not output the rest of the response.

  Returns:
    A live commentator processor.
  """
  detection_model = genai_model.GenaiModel(
      api_key=api_key,
      model_name=MODEL_DETECTION,
      generate_content_config=genai_types.GenerateContentConfig(
          system_instruction=EVENT_DETECTION_PROMPT,
          max_output_tokens=10,
          response_mime_type='text/x.enum',
          response_schema=EventTypes,
          media_resolution=MEDIA_RESOLUTION,
      ),
  )
  event_detection_processor = event_detection.EventDetection(
      backend=detection_model,
      output_dict={
          ('*', EventTypes.DETECTION): [
              content_api.ProcessorPart(
                  'start commentating',
                  role='user',
                  substream_name='realtime',
                  metadata={'turn_complete': True},
              )
          ],
          (EventTypes.DETECTION, EventTypes.NO_DETECTION): [
              content_api.ProcessorPart(
                  'stop commentating',
                  role='user',
                  substream_name='realtime',
                  metadata={'turn_complete': True},
              )
          ],
          (EventTypes.DETECTION, EventTypes.INTERRUPTION): [
              content_api.ProcessorPart(
                  '',
                  role='user',
                  # Setting up a substream name here will ensure this part will
                  # not be sent to the Live API.
                  substream_name='event_detection',
                  metadata={'interrupt_request': True},
              )
          ],
          (EventTypes.INTERRUPTION, EventTypes.DETECTION): None,
      },
      sensitivity={
          (
              EventTypes.DETECTION,
              EventTypes.NO_DETECTION,
          ): NO_DETECTION_SENSITIVITY,
      },
  )
  live_api_processor = live_model.LiveProcessor(
      api_key=api_key,
      model_name=MODEL_LIVE,
      realtime_config=genai_types.LiveConnectConfig(
          tools=TOOLS,
          system_instruction=PROMPT_PARTS,
          output_audio_transcription={},
          realtime_input_config=genai_types.RealtimeInputConfig(
              turn_coverage='TURN_INCLUDES_ALL_INPUT'
          ),
          response_modalities=['AUDIO'],
          generation_config=genai_types.GenerationConfig(
              media_resolution=MEDIA_RESOLUTION
          ),
      ),
      http_options=genai_types.HttpOptions(api_version='v1alpha'),
  )
  return (
      event_detection_processor
      + LiveCommentator(
          live_api_processor=live_api_processor,
          chattiness=chattiness,
          unsafe_string_list=unsafe_string_list,
      )
      + rate_limit_audio.RateLimitAudio(RECEIVE_SAMPLE_RATE)
  )



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""MCP server for examples.

This modules provides several MCP servers that can be used in CLI or examples:

*  `get_demo_mcp_session`: a self-contained MCP server. Good for tests, but
   doesn't do anything useful.

*  `get_local_mcp_session`: launches a given binary with an MCP server locally.

*  `get_remote_mcp_session`: connects to a remote MCP server. Will likely
   require an API key.
"""

from collections.abc import AsyncGenerator
import contextlib
import shlex
import httpx
import mcp
from mcp.client import session as mcp_session
from mcp.client import stdio
from mcp.client import streamable_http
from mcp.server import fastmcp
from mcp.shared import memory as mcp_memory


def _create_demo_server() -> fastmcp.FastMCP:
  """Create a demo MCP server with simple example tools."""
  server = fastmcp.FastMCP(name='Demo MCP Server')

  @server.tool()
  def add(a: int, b: int) -> int:
    """Add two numbers together.

    Args:
      a: The first number to add.
      b: The second number to add.

    Returns:
      The sum of the two numbers.
    """
    return a + b

  @server.tool()
  def multiply(a: int, b: int) -> int:
    """Multiply two numbers together.

    Args:
      a: The first number to multiply.
      b: The second number to multiply.

    Returns:
      The product of the two numbers.
    """
    return a * b

  @server.tool()
  def greet(name: str) -> str:
    """Greet someone by name.

    Args:
      name: The name of the person to greet.

    Returns:
      A greeting message.
    """
    return f'Hello, {name}! Nice to meet you.'

  @server.tool()
  def get_weather(city: str) -> str:
    """Get the current weather for a city (demo data).

    Args:
      city: The name of the city to get the weather for.

    Returns:
      The current weather for the specified city.
    """
    weather_data = {
        'london': 'Cloudy, 15°C',
        'paris': 'Sunny, 22°C',
        'tokyo': 'Rainy, 18°C',
        'new york': 'Partly cloudy, 20°C',
    }
    return weather_data.get(
        city.lower(), f'Weather data not available for {city}'
    )

  return server


@contextlib.asynccontextmanager
async def get_demo_mcp_session() -> (
    AsyncGenerator[mcp_session.ClientSession, None]
):
  """Returns a demo MCP server session.

  Returns:
    A demo MCP server session with a few example tools.
  """
  server = _create_demo_server()
  async with mcp_memory.create_connected_server_and_client_session(
      server
  ) as session:
    yield session


@contextlib.asynccontextmanager
async def get_local_mcp_session(
    server_command: str = 'npx -y @modelcontextprotocol/server-filesystem .',
) -> AsyncGenerator[mcp_session.ClientSession, None]:
  """Returns an external MCP server session.

  Args:
    server_command: The command to run the local MCP server.

  Returns:
    A local MCP server session.
  """
  parts = shlex.split(server_command)
  server_params = stdio.StdioServerParameters(command=parts[0], args=parts[1:])
  print(f'Connecting to MCP server: {server_command}')
  async with stdio.stdio_client(server_params) as (read, write):
    async with mcp_session.ClientSession(read, write) as session:
      await session.initialize()
      yield session


@contextlib.asynccontextmanager
async def get_remote_mcp_session(
    server_address: str = 'https://mapstools.googleapis.com/mcp',
    api_key_header: dict[str, str] | None = None,
) -> AsyncGenerator[mcp_session.ClientSession, None]:
  """Returns a remote MCP server session.

  Args:
    server_address: The address of the remote MCP server.
    api_key_header: dictionary of headers containing the API key for the server.
      E.g. { 'X-Goog-Api-Key': <google_api_key> }.

  Returns:
    A remote MCP server session.
  """
  headers = {
      'Content-Type': 'application/json',
      'Accept': 'application/json, text/event-stream',
  }
  if api_key_header:
    headers.update(api_key_header)
  async with httpx.AsyncClient(
      headers=headers,
      timeout=httpx.Timeout(30, read=60),
  ) as client:
    async with streamable_http.streamable_http_client(
        server_address, http_client=client
    ) as (read, write, session_id):
      async with mcp.ClientSession(read, write) as session:
        print(f'Connecting to MCP server: {server_address}')
        await session.initialize()
        yield session



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processor which researches a `Topic`, using tools."""

from typing import AsyncIterable

from genai_processors import content_api
from genai_processors import processor
from genai_processors.core import genai_model
from genai_processors.core import jinja_template
from genai_processors.core import preamble
from google.genai import types

from .. import interfaces
from .. import prompts

ProcessorPart = processor.ProcessorPart


class TopicResearcher(processor.PartProcessor):
  """Processor which researches a given `Topic`.

  This processor uses the GenAI API's tool feature to research a given
  topic, and yields `TopicResearch` ProcessorParts with the research results.
  """

  def __init__(
      self,
      api_key: str,
      config: interfaces.Config | None = None,
  ):
    """Initializes the TopicResearcher.

    Args:
      api_key: The API key to use for the GenAI API.
      config: The agent configuration.
    """
    self._config = config or interfaces.Config()
    self._genai_processor = genai_model.GenaiModel(
        api_key=api_key,
        model_name=self._config.topic_researcher_model_name,
        generate_content_config=types.GenerateContentConfig(
            tools=self._config.enabled_research_tools,
        ),
        http_options=types.HttpOptions(
            retry_options=types.HttpRetryOptions(
                attempts=100,
            ),
        ),
    )
    p_preamble = preamble.Preamble(
        content=[
            ProcessorPart(prompts.TOPIC_RESEARCH_PREAMBLE),
            ProcessorPart('Topic to research: '),
        ]
    )
    p_verbalizer = jinja_template.RenderDataClass(
        template_str=(
            '## {{ data.topic }}\n'
            '*{{ data.relationship_to_user_content }}*'
            '{% if data.research_text|trim != "" %}'
            '\n\n### Research\n\n{{ data.research_text }}'
            '{% endif %}'
        ),
        data_class=interfaces.Topic,
    )
    p_suffix = preamble.Suffix(content=[ProcessorPart('Your research: ')])
    self._pipeline = (
        p_verbalizer + p_preamble + p_suffix + self._genai_processor
    )

  def match(self, part: ProcessorPart) -> bool:
    return content_api.is_dataclass(part.mimetype, interfaces.Topic)

  async def call(
      self,
      part: ProcessorPart,
  ) -> AsyncIterable[ProcessorPart]:

    input_topic = part.get_dataclass(interfaces.Topic)
    response = self._pipeline(part)

    updated_topic = interfaces.Topic(
        topic=input_topic.topic,
        relationship_to_user_content=input_topic.relationship_to_user_content,
        research_text=await response.text(),
    )
    yield ProcessorPart.from_dataclass(updated_topic)
    yield processor.status(f"""Researched topic!

## {updated_topic.topic}

### Research

{updated_topic.research_text}""")



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Command Line Interface to run a simple live audio agent.

This agent uses the Gemini API with live/conversation.py to use a turn-based
Gemini model to interact with the user in a real-time conversation similar to
a Live API setup. This is audio only.

## Setup

To install the dependencies for this script, run:

```
pip install --upgrade pyaudio genai-processors google-genai termcolor \
    langchain-google-genai transformers google-cloud-texttospeech
```

Note that you need to activate the Google Text-to-Speech API in your Google
Cloud project (as well as speech to text).

Before running this script, ensure the `GOOGLE_API_KEY` and `GOOGLE_PROJECT_ID`
environment variables are set to the api-key you obtained from Google AI Studio.

Important: **Use headphones**. This script uses the system default audio
input and output, which often won't include echo cancellation. So to prevent
the model from interrupting itself it is important to use headphones.

## Run

To run the script:

```shell
python3 ./realtime_simple_cli.py
```
"""

import asyncio
import os
from typing import Sequence

from absl import app
from genai_processors import content_api
from genai_processors import context
from genai_processors import processor
from genai_processors.core import audio_io
from genai_processors.core import rate_limit_audio
from genai_processors.core import realtime
from genai_processors.core import speech_to_text
from genai_processors.core import text
from genai_processors.core import text_to_speech
from genai_processors.examples import models
import pyaudio

# You need to define the project id in the environment variables.
GOOGLE_PROJECT_ID = os.environ['GOOGLE_PROJECT_ID']

INSTRUCTION_PARTS = [
    'You are an agent that interacts with the user in a conversation. Make'
    ' the conversation lively and interesting for the user. You can make jokes,'
    ' explain interesting facts related to what you see and hear, predict what'
    ' could happen, judge some actions or reactions, etc. Respond to the'
    ' user in a few sentences maximum: keep it short and engaging. Avoid'
    ' long monologues. You can use Google search to add extra information to'
    ' the user questions or to come up with interesting news or facts.'
]


@processor.create_filter
def _filter_parts(part: content_api.ProcessorPart) -> bool:
  """Filters out parts that are not relevant to a conversation.

  Removes audio responses from previous turns (model only) and reserved
  substreams. This is needed when working with an LLM that takes text only.
  It is also recommended (but not required) for Audio-In models to avoid
  re-tokenizing the same audio again and to use the transcription as input
  instead.

  Args:
    part: the part to filter.

  Returns:
    True if this filter should be passed through, False if it is filtered out.
  """
  if context.is_reserved_substream(part.substream_name):
    return False
  # Filters out audio responses from previous turns (model only)
  if content_api.is_audio(part.mimetype) and part.role.lower() == 'model':
    return False
  return True


async def run_conversation() -> None:
  r"""Runs a simple conversation agent taking an audio stream as input.

  The audio input and output is connected to the local machine's default input
  and output devices.
  """

  # input processor = audio stream from the default input device/mic + STT
  # The STT processor is used to convert the audio stream to text. This will
  # be used to store the conversation history in the prompt.
  pya = pyaudio.PyAudio()
  input_processor = audio_io.PyAudioIn(pya) + speech_to_text.SpeechToText(
      project_id=GOOGLE_PROJECT_ID,
      with_interim_results=False,
  )

  # Main model that will be used to generate the response. Note that filter
  # before the genai model that will remove the audio parts and will make sure
  # only text is sent to the model.
  genai_processor = _filter_parts + models.turn_based_model(
      system_instruction=INSTRUCTION_PARTS
  )

  # TTS processor that will be used to convert the text response to audio. Note
  # the rate limit audio processor that will be used to stream back small audio
  # chunks to the client at the same rate as how they are played back. This is
  # needed to stop the audio when the user is speaking: the rate limit audio
  # processor will then stop and the audio will not be played anymore.
  tts = text_to_speech.TextToSpeech(
      project_id=GOOGLE_PROJECT_ID
  ) + rate_limit_audio.RateLimitAudio(
      sample_rate=24000,
      delay_other_parts=True,
  )

  # Plays the audio parts. This processor also handles interruptions and makes
  # sure the audio output stops when the user is speaking.
  play_output = audio_io.PyAudioOut(pya)

  # Creates an agent as:
  # mic -> speech to text -> text conversation -> text to speech -> play audio
  conversation_agent = (
      input_processor
      + realtime.LiveProcessor(turn_processor=genai_processor + tts)
      + play_output
  )
  prompt = 'USER (ctrl+D to end)> '
  await text.terminal_output(
      conversation_agent(text.terminal_input(prompt=prompt)), prompt=prompt
  )


def main(argv: Sequence[str]):
  del argv  # Unused.
  if not GOOGLE_PROJECT_ID:
    raise ValueError(
        'Project ID is not set. Define a GOOGLE_PROJECT_ID environment variable'
        ' obtained from your Cloud project.'
    )
  asyncio.run(run_conversation())


if __name__ == '__main__':
  app.run(main)



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Wraps the Gemini Live API into a Processor.

The list of available models can be found at
https://ai.google.dev/gemini-api/docs/models
Note that only Live models will work with the LiveProcessor.

## Example usage

```py
p = LiveProcessor(
    api_key=API_KEY,
    model_name="gemini-2.0-flash-live-001",
)
```

Given a stream of content, you can use the processor to generate
a stream of content as follows:

```py
async for part in p(['hello', 'world']):
  # Do something with part
```

The Live Processor processor only considers parts with the substream name
"realtime" as input (sent to real-time methods), or with the default substream
name (sent to content generate method).
"""

import asyncio
from collections.abc import AsyncIterable
import re
import time
from typing import Iterable, Optional
from absl import logging
from genai_processors import content_api
from genai_processors import processor
from google.genai import client
from google.genai import types as genai_types


PartTypes = content_api.ProcessorPartTypes
ProcessorPart = content_api.ProcessorPart


def to_parts(
    msg: genai_types.LiveServerMessage,
) -> Iterable[content_api.ProcessorPart]:
  """Converts a LiveServerMessage to a stream of ProcessorParts."""
  if msg.server_content:
    metadata = msg.server_content.to_json_dict()
    if 'model_turn' in metadata:
      del metadata['model_turn']
    if msg.server_content.model_turn:
      for part in msg.server_content.model_turn.parts:
        yield content_api.ProcessorPart(
            value=part,
            role=msg.server_content.model_turn.role,
        )
    for k, v in metadata.items():
      value = ''
      if k in ('input_transcription', 'output_transcription'):
        if 'text' in v:
          value = v['text']
        yield content_api.ProcessorPart(
            value=value,
            role='model',
            substream_name=k,
        )
      else:
        yield content_api.ProcessorPart(
            value='',
            role='model',
            metadata={k: v},
        )
  if msg.tool_call:
    function_calls = msg.tool_call.function_calls
    for function_call in function_calls:
      yield content_api.ProcessorPart.from_function_call(
          name=function_call.name,
          args=function_call.args,
          role='model',
          metadata={'id': function_call.id},
      )
  if msg.tool_call_cancellation and msg.tool_call_cancellation.ids:
    for function_call_id in msg.tool_call_cancellation.ids:
      yield content_api.ProcessorPart.from_tool_cancellation(
          function_call_id=function_call_id,
      )
  if msg.usage_metadata:
    yield content_api.ProcessorPart(
        value='',
        role='model',
        metadata={'usage_metadata': msg.usage_metadata.to_json_dict()},
    )

  if msg.go_away:
    yield content_api.ProcessorPart(
        value='',
        role='model',
        metadata={'go_away': msg.go_away.to_json_dict()},
    )

  if msg.session_resumption_update:
    yield content_api.ProcessorPart(
        value='',
        role='model',
        metadata={
            'session_resumption_update': (
                msg.session_resumption_update.to_json_dict()
            )
        },
    )


class LiveProcessor(processor.Processor):
  """Gemini Live API Processor to generate realtime content.

  The realtime content captured via mic and camera should be passed to the
  processor with the `realtime` substream name. The default substream is used
  for standard content input.

  An image sent on the default substream will be processed by the model as an
  ad-hoc user input, not as a realtime input captured from realtime devices.
  This lets the user send an image to the model and ask a question about it,
  for example "What is this?", independently of the video stream being sent to
  the model on the `realtime` substream.
  """

  def __init__(
      self,
      api_key: str,
      model_name: str,
      realtime_config: Optional[genai_types.LiveConnectConfigOrDict] = None,
      debug_config: client.DebugConfig | None = None,
      http_options: (
          genai_types.HttpOptions | genai_types.HttpOptionsDict | None
      ) = None,
  ):
    """Initializes the Live Processor.

    Args:
      api_key: The [API key](https://ai.google.dev/gemini-api/docs/api-key) to
        use for authentication. Applies to the Gemini Developer API only.
      model_name: The name of the model to use. See
        https://ai.google.dev/gemini-api/docs/models for a list of available
          models. Only use models with a `-live-` suffix.
      realtime_config: The configuration for generating realtime content.
      debug_config: Config settings that control network behavior of the client.
        This is typically used when running test code.
      http_options: Http options to use for the client. These options will be
        applied to all requests made by the client. Example usage: `client =
        genai.Client(http_options=types.HttpOptions(api_version='v1'))`.

    Returns:
      A `Processor` that calls the Genai API in a realtime (aka live) fashion.
    """
    self._client = client.Client(
        api_key=api_key,
        debug_config=debug_config,
        http_options=http_options,
    )
    self._model_name = model_name
    self._realtime_config = realtime_config

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:

    output_queue = asyncio.Queue[Optional[ProcessorPart]](maxsize=1_000)

    async with self._client.aio.live.connect(
        model=self._model_name,
        config=self._realtime_config,
    ) as session:

      async def consume_content():
        async for chunk_part in content:
          if chunk_part.part.function_response:
            logging.debug(
                '%s - Live Processor: sending tool response: %s',
                time.perf_counter(),
                chunk_part,
            )
            await session.send_tool_response(
                function_responses=chunk_part.part.function_response
            )
          elif (
              chunk_part.substream_name == 'realtime'
              and chunk_part.get_metadata('audio_stream_end')
          ):
            logging.debug(
                '%s - Live Processor: sending realtime audio_stream_end',
                time.perf_counter(),
            )
            await session.send_realtime_input(audio_stream_end=True)
          elif (
              chunk_part.substream_name == 'realtime'
              and chunk_part.part.inline_data
          ):
            await session.send_realtime_input(media=chunk_part.part.inline_data)
          elif chunk_part.substream_name == 'realtime' and content_api.is_text(
              chunk_part.mimetype
          ):
            logging.debug(
                '%s - Live Processor: sending realtime input: %s',
                time.perf_counter(),
                chunk_part.text,
            )
            await session.send_realtime_input(text=chunk_part.text)
          elif not chunk_part.substream_name:
            # Default substream.
            logging.debug(
                '%s - Live Processor: sending client content: %s',
                time.perf_counter(),
                chunk_part.part,
            )
            turn_complete = chunk_part.get_metadata('turn_complete')
            await session.send_client_content(
                turns=genai_types.Content(
                    parts=[chunk_part.part], role=chunk_part.role
                ),
                turn_complete=True if turn_complete is None else turn_complete,
            )
          else:
            logging.debug(
                '%s - Live Processor: part passed through: %s',
                time.perf_counter(),
                chunk_part,
            )
            await output_queue.put(chunk_part)
        await output_queue.put(None)

      async def produce_content():
        try:
          while True:
            async for response in session.receive():
              if not (
                  response.server_content
                  and response.server_content.model_turn
                  and response.server_content.model_turn.parts
                  and response.server_content.model_turn.parts[0].inline_data
              ):
                logging.debug(
                    '%s - Live Processor Response: %s',
                    time.perf_counter(),
                    # Remove the None values from the response.
                    re.sub(r'(,\s)?[^\(\s]+=None,?\s?', '', str(response)),
                )
              for part in to_parts(response):
                await output_queue.put(part)
            # Allow `yield` if session.receive() does not return anything.
            await asyncio.sleep(0)
        finally:
          await output_queue.put(None)

      consume_content_task = processor.create_task(consume_content())
      produce_content_task = processor.create_task(produce_content())

      while chunk := await output_queue.get():
        yield chunk

      consume_content_task.cancel()
      produce_content_task.cancel()



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Examples are in the root of the git repository, see the parent folder.

To make examples easily discoverable we need them to be in the parent folder.
But when doing local development with `PYTHONPATH=.`Python will look for them
here. This magic tells Python where the examples packages are really located.
"""

import importlib
import sys


class ExamplesImporter(object):

  def find_spec(self, fullname: str, path: str, target=None):
    del path, target  # Unused.
    if fullname.startswith('genai_processors.examples'):
      return importlib.util.find_spec(fullname.replace('genai_processors.', ''))
    return None


sys.meta_path.append(ExamplesImporter())



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""ADK - GenAI Processors integration."""

from typing import AsyncGenerator, AsyncIterable, Callable
from typing_extensions import override

from genai_processors import content_api
from genai_processors import processor
from google.adk.agents import base_agent
from google.adk.agents import invocation_context
from google.adk.events import event as adk_event
from google.adk.events import event_actions
from google.genai import types as genai_types


class ProcessorAgent(base_agent.BaseAgent):
  """ADK Custom agent that delegates processing its input to a Processor.

  Works both for turn-based and live modes. In case of turn-based mode the
  Processor will be called for each user turn with the whole conversation
  history up to that turn.

  This agent does use ADK token streaming. "Token streaming" checkbox in the ADK
  UI must be enabled for the response to be rendered correctly. If used
  programmatically, consumer must either only use events marked as partial=True
  or turn_complete=True, but not both.
  """

  def __init__(
      self, processor_factory: Callable[[], processor.Processor], *, name: str
  ):
    """Initializes the ProcessorAgent.

    Args:
      processor_factory: A function that returns a Processor to be applied to
        the incoming content. It will be called on each turn and each request.
        Unless the returned processor is stateless, it must return a new
        instance every time to avoid state sharing between users.
      name: The agent's name. It must be a valid Python identifier and unique
        within the agent tree. It can't be "user", since it's reserved for
        end-user's input.
    """
    super().__init__(name=name)
    self._processor_factory = processor_factory

  def _append_to_history(
      self,
      ctx: invocation_context.InvocationContext,
      content: genai_types.Content,
  ) -> adk_event.Event:
    # Parsing conversation history from the Event log requires handling many
    # edge cases which ADK considers to be implementation details. Currently
    # only adk.LlmAgent is priviledged to do that. As a temporary workaround we
    # will accumulate the history in the state. Downside is that event log
    # (stored in memory) will grow as n^2.
    key = f'history_{self.name}'
    history = ctx.session.state.get(key, [])
    history.append(content)
    return adk_event.Event(
        actions=event_actions.EventActions(state_delta={key: history}),
        author=self.name,
    )

  async def _stream_history(
      self, ctx: invocation_context.InvocationContext
  ) -> AsyncIterable[content_api.ProcessorPart]:
    for content in ctx.session.state[f'history_{self.name}']:
      for part in content.parts:
        yield content_api.ProcessorPart(part, role=content.role)

  @override
  async def _run_async_impl(
      self, ctx: invocation_context.InvocationContext
  ) -> AsyncGenerator[adk_event.Event, None]:
    p = self._processor_factory()

    yield self._append_to_history(ctx, ctx.user_content)
    response = genai_types.Content(parts=[], role='model')
    async for part in p(self._stream_history(ctx)):
      yield adk_event.Event(
          content=genai_types.Content(parts=[part.part], role='model'),
          author=self.name,
          partial=True,
          invocation_id=ctx.invocation_id,
      )
      response.parts.append(part.part)

    final_event = self._append_to_history(ctx, response)
    final_event.content = response
    final_event.turn_complete = True
    yield final_event

  @override
  async def _run_live_impl(
      self, ctx: invocation_context.InvocationContext
  ) -> AsyncGenerator[adk_event.Event, None]:
    async def stream_content():
      while True:
        request = await ctx.live_request_queue.get()
        if request.blob:
          yield content_api.ProcessorPart(
              request.blob.data,
              mimetype=request.blob.mime_type,
              substream_name='realtime',
              role='user',
          )
        if request.close:
          # NOTE: Currently ADK Web doesn't close connection and this case is
          # unreachable. _run_live_impl handlers will leak. Fixing this is in
          # ADK backlog.
          break

    p = self._processor_factory()
    async for part in p(stream_content()):
      if not part.role:
        part.role = 'model'
      yield adk_event.Event(
          content=genai_types.Content(parts=[part.part]), author=self.name
      )



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processors for extracting file content from Google Drive content.

## Example usage

### Authentication

Auth is handled either by "pre-authorizing" in Colab, e.g.:

```py
from google.colab import auth as colab_auth
colab_auth.authenticate_user()
```

Or, you will need to provide credentials directly to the processor in the
constructor, e.g.:

```py
from google.auth import credentials as google_credentials

creds = google_credentials.Credentials(
    token=YOUR_ACCESS_TOKEN,
    scopes=['https://www.googleapis.com/auth/drive']
)
p_slides = drive.Slides(creds=creds)
```

### PDF Handling

In the examples below, we pass the PDF bytes directly to the `GenaiModel`,
since the `GenaiModel` will handle the PDF conversion.

If you require more control over the PDF conversion, you could make use of
the `PDFExtract` processor.

## Docs Example

```py
from genai_processors.core import drive

USER_PROMPT = 'Describe the document in detail.'
p_docs = drive.Docs()

# Or, if you are using credentials:
# p_docs = drive.Docs(creds=creds)

p_preamble = preamble.Preamble(
    content=f'''You are an expert in reviewing Google Docs documents.

You have been provided with a document, and must use it to answer the question.

User question: {USER_PROMPT}'''
)

p_genai = genai_model.GenaiModel(
    model_name='gemini-2.5-flash',
    api_key=GOOGLE_API_KEY
)

pipeline = (
    p_docs + p_preamble + p_genai
)

req = drive.DocsRequest(doc_id='YOUR-DOC-ID')

req_part = processor.ProcessorPart.from_dataclass(req)

input_stream = processor.stream_content([req_part])

async for content_part in pipeline(input_stream):
  print(content_part.text)
```

## Sheets Example

```py
from genai_processors.core import drive

USER_PROMPT = "Describe the spreadsheet in detail."

p_sheets = drive.Sheets()

# Or, if you are using credentials:
# p_sheets = drive.Sheets(creds=creds)

p_preamble = preamble.Preamble(
    content=f'''You are an expert in reviewing Google Sheets spreadsheets.

You have been provided with a spreadsheet, and must use it to answer the
question.

User question: {USER_PROMPT}'''
)

p_genai = genai_model.GenaiModel(
    model_name='gemini-2.5-flash',
    api_key=GOOGLE_API_KEY
)

pipeline = (
    p_sheets + p_preamble + p_genai
)

req = drive.SheetsRequest(spreadsheet_id='YOUR-SPREADSHEET-ID')

req_part = processor.ProcessorPart.from_dataclass(req)

input_stream = processor.stream_content([req_part])

async for content_part in pipeline(input_stream):
  print(content_part.text)
```

## Slides Example

```py
from genai_processors.core import drive

USER_PROMPT = "Describe the presentation in detail."

p_slides = drive.Slides()

# Or, if you are using credentials:
# p_slides = drive.Slides(creds=creds)

p_preamble = preamble.Preamble(
    content=f'''You are an expert in reviewing Google Slides presentations.

You have been provided with slides, and must use them to answer the question.

User question: {USER_PROMPT}'''
)

p_genai = genai_model.GenaiModel(
    model_name='gemini-2.5-flash',
    api_key=GOOGLE_API_KEY
)

pipeline = (
    p_slides + p_preamble + p_genai
)

req = drive.SlidesRequest(presentation_id='YOUR-PRESENTATION-ID')

req_part = processor.ProcessorPart.from_dataclass(req)

input_stream = processor.stream_content([req_part])

async for content_part in pipeline(input_stream):
  print(content_part.text)
```
"""

import csv
import dataclasses
import io
from typing import Any, AsyncIterable
import dataclasses_json
from genai_processors import content_api
from genai_processors import processor
from google.auth import credentials as google_credentials
from googleapiclient import discovery
import pdfrw

# Shared helper functions


def get_drive_pdf(
    file_id: str,
    creds: google_credentials.Credentials | None = None,
) -> bytes:
  """Fetches the requested file as PDF bytes.

  Args:
    file_id: The ID of the file to fetch.
    creds: The credentials to use for the request. If not provided, the
      application default credentials will be used.

  Returns:
    The requested file as PDF bytes.
  """
  service = discovery.build('drive', 'v3', credentials=creds)
  return (
      service.files()
      .export(fileId=file_id, mimeType='application/pdf')
      .execute()
  )


# Google Docs
@dataclasses_json.dataclass_json
@dataclasses.dataclass
class DocsRequest:
  """Request for fetching Google Docs data.

  Attributes:
    doc_id: The ID of the document to fetch. (e.g. for a doc with URL
      https://docs.google.com/document/d/foo, the ID is "foo")
  """

  doc_id: str


class Docs(processor.PartProcessor):
  """Processor for extracting PDF content from Google Docs."""

  def __init__(
      self,
      creds: google_credentials.Credentials | None = None,
  ) -> None:
    self._creds = creds

  def match(self, part: content_api.ProcessorPart) -> bool:
    return part.mimetype == 'application/json; type=DocsRequest'

  async def call(
      self, part: content_api.ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPart]:
    google_docs_request = part.get_dataclass(DocsRequest)
    doc_id = google_docs_request.doc_id
    doc_pdf = get_drive_pdf(
        file_id=doc_id,
        creds=self._creds,
    )
    yield content_api.ProcessorPart('Document:\n\n')
    yield content_api.ProcessorPart(doc_pdf, mimetype='application/pdf')


# Google Sheets
@dataclasses_json.dataclass_json
@dataclasses.dataclass
class SheetsRequest:
  """Request for fetching Google Sheets data.

  Attributes:
    spreadsheet_id: The ID of the spreadsheet to fetch. (e.g. for a spreadsheet
      with URL https://docs.google.com/spreadsheets/d/foo, the ID is "foo")
    ranges: Cell ranges to fetch. If not specified, the entire worksheet is
      fetched. For additional details on how to specify the ranges, see here the
      `ranges` paramter in
      https://developers.google.com/workspace/sheets/api/reference/rest/v4/spreadsheets/get
    worksheet_names: The names of the worksheets to fetch. If not provided, all
      worksheets will be fetched.
  """

  spreadsheet_id: str
  ranges: list[str] | None = None
  worksheet_names: list[str] | None = None


class Sheets(processor.PartProcessor):
  """Processor for extracting PDF content from Google Sheets."""

  def __init__(
      self,
      creds: google_credentials.Credentials | None = None,
  ) -> None:
    self._creds = creds

  def _fetch_sheet_data(
      self, sheets_request: SheetsRequest
  ) -> list[dict[str, Any]]:
    """Returns a list of dicts representing sheet data, for the given spreadsheet or range.

    The response dicts contain sheets data with the fields described here:
    https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/sheets

    Args:
      sheets_request: The SheetsRequest to fetch.

    Returns:
      A list of response dicts for the given spreadsheet or range.
    """
    service = discovery.build('sheets', 'v4', credentials=self._creds)
    spreadsheet_id = sheets_request.spreadsheet_id
    ranges = sheets_request.ranges
    return [
        service.spreadsheets()
        .get(spreadsheetId=spreadsheet_id, includeGridData=True, ranges=r)
        .execute()
        for r in ranges or [None]
    ]

  def match(self, part: content_api.ProcessorPart) -> bool:
    return part.mimetype == 'application/json; type=SheetsRequest'

  async def call(
      self, part: content_api.ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPart]:
    google_sheets_request = part.get_dataclass(SheetsRequest)
    spreadsheet_data = self._fetch_sheet_data(google_sheets_request)
    sheet_ranges = google_sheets_request.ranges
    worksheet_names = google_sheets_request.worksheet_names
    for i, res in enumerate(spreadsheet_data):
      for sheet in res['sheets']:
        try:
          title = sheet['properties']['title']
          if worksheet_names is not None and title not in worksheet_names:
            continue
          data = []
          for row in sheet['data'][0]['rowData']:
            if 'values' in row.keys():
              values = []
              for cell in row['values']:
                values.append(cell.get('formattedValue', ''))
              data.append(values)
          columns = data[0]
          num_columns = len(columns)
          rows = data[1:]
          for index in range(len(rows)):
            if len(rows[index]) > len(columns):
              rows[index] = rows[index][:num_columns]
            if len(rows[index]) < len(columns):
              rows[index] = rows[index] + [''] * (
                  num_columns - len(rows[index])
              )
          output = io.StringIO()
          writer = csv.writer(output, lineterminator='\n')
          writer.writerow(columns)
          writer.writerows(rows)
          range_or_title = sheet_ranges[i] if sheet_ranges else title
          yield content_api.ProcessorPart(f'Sheet {range_or_title}:\n\n')
          yield content_api.ProcessorPart(
              output.getvalue(), mimetype='text/csv'
          )
        except (ValueError, TypeError, IndexError):
          yield content_api.ProcessorPart('Failed to parse sheet data.')


# Google Slides
@dataclasses_json.dataclass_json
@dataclasses.dataclass
class SlidesRequest:
  """Request for fetching Google Slides data.

  Attributes:
    presentation_id: The ID of the presentation to fetch. (e.g. for a
      presentation with URL https://docs.google.com/presentation/d/foo, the ID
      is "foo")
    slide_numbers: The slide numbers to fetch, indexed from 1. If not provided,
      all slides will be fetched.
  """

  presentation_id: str
  slide_numbers: list[int] | None = None


class Slides(processor.PartProcessor):
  """Processor for extracting PDF content from Google Slides."""

  def __init__(
      self,
      creds: google_credentials.Credentials | None = None,
  ) -> None:
    self._creds = creds

  def _get_presentation_pdf_by_slide(
      self,
      presentation_id: str,
      slide_numbers: list[int] | None = None,
  ) -> dict[int, bytes]:
    """Fetches the requested presentation as a dictionary of slide numbers to PDF bytes.

    Args:
      presentation_id: The ID of the presentation to fetch.
      slide_numbers: (Optional) A list of slide numbers to fetch (indexed from
        1). If not provided, all slides will be fetched.

    Returns:
      A dictionary of slide numbers to PDF bytes.
    """
    pdf_bytes = get_drive_pdf(
        file_id=presentation_id,
        creds=self._creds,
    )
    read_pdf = pdfrw.PdfReader(io.BytesIO(pdf_bytes))
    num_slides = read_pdf.numPages
    slide_dict = {}
    for i in range(num_slides):
      if slide_numbers and i + 1 not in slide_numbers:
        continue
      slide_writer = pdfrw.PdfWriter()
      slide_writer.addPage(read_pdf.getPage(i))
      slide_bytes_stream = io.BytesIO()
      slide_writer.write(slide_bytes_stream)
      slide_bytes_stream.seek(0)
      slide_bytes = slide_bytes_stream.read()
      slide_dict[i + 1] = slide_bytes
    return slide_dict

  def match(self, part: content_api.ProcessorPart) -> bool:
    return part.mimetype == 'application/json; type=SlidesRequest'

  async def call(
      self, part: content_api.ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPart]:
    google_slides_request = part.get_dataclass(SlidesRequest)
    presentation_id = google_slides_request.presentation_id
    slide_numbers = google_slides_request.slide_numbers
    presentation_pdfs = self._get_presentation_pdf_by_slide(
        presentation_id=presentation_id,
        slide_numbers=slide_numbers,
    )
    for slide_num, pdf_bytes in presentation_pdfs.items():
      yield content_api.ProcessorPart(f"""Slide {slide_num}:\n\n""")
      yield content_api.ProcessorPart(pdf_bytes, mimetype='application/pdf')
